/* Core MMC driver functions
 *
 * Copyright (c) 2002 Hewlett-Packard Company
 *   
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *  
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Many thanks to Alessandro Rubini and Jonathan Corbet!
 *
 * Author:  Andrew Christian
 *          6 May 2002 
 */

/*
 * mmc_protocol.c - MMC protocol driver
 *
 * Copyright (C) 2006 Intel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/sysctl.h>
#include <linux/mm.h>
#include <asm/scatterlist.h>

#include <linux/mmc/mss_core.h>
#include <linux/mmc/mmc_protocol.h>

#if 0
static char * mmc_result_strings[] = {
	"NO_ERROR",
	"ERROR_OUT_OF_RANGE",
	"ERROR_ADDRESS",
	"ERROR_BLOCK_LEN",
	"ERROR_ERASE_SEQ",
	"ERROR_ERASE_PARAM",
	"ERROR_WP_VIOLATION",
	"ERROR_CARD_IS_LOCKED",
	"ERROR_LOCK_UNLOCK_FAILED",
	"ERROR_COM_CRC",
	"ERROR_ILLEGAL_COMMAND",
	"ERROR_CARD_ECC_FAILED",
	"ERROR_CC",
	"ERROR_GENERAL",
	"ERROR_UNDERRUN",
	"ERROR_OVERRUN",
	"ERROR_CID_CSD_OVERWRITE",
	"ERROR_STATE_MISMATCH",
	"ERROR_HEADER_MISMATCH",
};

static char * mmc_state_strings[] = {
	"empty",
	"idle",
	"ready",
	"ident",
	"stby",
	"tran",
	"data",
	"rcv",
	"prg",
	"dis",
};

/*****************************************************************************
 *
 *   internal functions
 *
 ****************************************************************************/

static inline char * mmc_result_to_string( int i )
{
	return mmc_result_strings[i+1];
}

static inline char * mmc_state_to_string( int i )
{
	return mmc_state_strings[i+1];
}
#endif

#define KBPS 1
#define MBPS 1000

static u32 ts_exp[] = { 100*KBPS, 1*MBPS, 10*MBPS, 100*MBPS, 0, 0, 0, 0 };
static u32 ts_mul[] = { 0,    1000, 1200, 1300, 1500, 2000, 2500, 3000, 
			3500, 4000, 4500, 5000, 5500, 6000, 7000, 8000 };

static u32 mmc_tran_speed(u8 ts)
{
	u32 clock = ts_exp[(ts & 0x7)] * ts_mul[(ts & 0x78) >> 3];

	return clock;
}
#if 0
static enum mss_result configure_card(struct mss_card_device *dev)
{
	struct mmc_card_info *card;
	u32 rate = 0;

	dbg("card type: %d, card: 0x%p", dev->card_type, dev);
	if (dev->card_type != MMC_CARD)
		return MSS_WRONG_CARD;

	/* Fix the clock rate */
	card = (struct mmc_card_info *)dev->card;
	dbg("mmc_card: 0x%p", card);
	rate = mmc_tran_speed(card->card_csd.tran_speed);
	if ( rate < CARD_CLOCK_SLOW )
		rate = CARD_CLOCK_SLOW;
	if ( rate > CARD_CLOCK_FAST )
		rate = CARD_CLOCK_FAST;
	dev->slot->controller->set_clock(dev, rate);
	return MSS_SUCCESS;
}
#endif

static int mmc_unpack_r1(struct mss_cmd *cmd, struct mmc_response_r1 *r1, struct mmc_card *mmc_card)
{
	u8 *buf = cmd->response;

	//debug(" result in r1: %d\n", request->result);
	//if ( request->result )  return request->result;

	mmc_card->errno = MMC_ERROR_NONE;
	r1->cmd    = buf[0];
	r1->status = PARSE_U32(buf + 1);

	if (R1_STATUS(r1->status)) {
		if (r1->status & R1_OUT_OF_RANGE)
			mmc_card->errno = MMC_ERROR_OUT_OF_RANGE;
		if (r1->status & R1_ADDRESS_ERROR)
		  	mmc_card->errno = MMC_ERROR_ADDRESS;
		if (r1->status & R1_BLOCK_LEN_ERROR)
		    	mmc_card->errno = MMC_ERROR_BLOCK_LEN;
		if (r1->status & R1_ERASE_SEQ_ERROR)
		    	mmc_card->errno = MMC_ERROR_ERASE_SEQ;
		if (r1->status & R1_ERASE_PARAM)
			mmc_card->errno = MMC_ERROR_ERASE_PARAM;
		if (r1->status & R1_WP_VIOLATION)
		 	mmc_card->errno = MMC_ERROR_WP_VIOLATION;
		if (r1->status & R1_LOCK_UNLOCK_FAILED)
			mmc_card->errno = MMC_ERROR_LOCK_UNLOCK_FAILED;
		if (r1->status & R1_COM_CRC_ERROR)
		  	mmc_card->errno = MMC_ERROR_COM_CRC;
		if (r1->status & R1_ILLEGAL_COMMAND)
		    	mmc_card->errno = MMC_ERROR_ILLEGAL_COMMAND;
		if (r1->status & R1_CARD_ECC_FAILED)
		    	mmc_card->errno = MMC_ERROR_CARD_ECC_FAILED;
		if (r1->status & R1_CC_ERROR)
	     		mmc_card->errno = MMC_ERROR_CC;
		if (r1->status & R1_ERROR)
	  		mmc_card->errno = MMC_ERROR_GENERAL;
		if (r1->status & R1_UNDERRUN)
	     		mmc_card->errno = MMC_ERROR_UNDERRUN;
		if (r1->status & R1_OVERRUN)
      			mmc_card->errno = MMC_ERROR_OVERRUN;
		if (r1->status & R1_CID_CSD_OVERWRITE)
		      	mmc_card->errno = MMC_ERROR_CID_CSD_OVERWRITE;
	}

	if (buf[0] != cmd->opcode)
	       	mmc_card->errno = MMC_ERROR_HEADER_MISMATCH;
	/* This should be last - it's the least dangerous error */
	if (R1_CURRENT_STATE(r1->status) != mmc_card->state ) {
		dbg5("state dismatch:r1->status:%x,state:%x\n",R1_CURRENT_STATE(r1->status),mmc_card->state);
		mmc_card->errno = MMC_ERROR_STATE_MISMATCH;
	}
	dbg5("mmc card error %d", mmc_card->errno);
	if (mmc_card->errno)
		return MSS_ERROR_RESP_UNPACK;
	return 0;
}

static int mmc_unpack_r3(struct mss_cmd *cmd, struct mmc_response_r3 *r3, struct mmc_card *mmc_card)
{
	u8 *buf = cmd->response;

	mmc_card->errno = MMC_ERROR_NONE;
	r3->ocr = PARSE_U32(buf + 1);
	//debug(" ocr=%08x\n", r3->ocr);

	if (buf[0] != 0x3f) {  
		mmc_card->errno = MMC_ERROR_HEADER_MISMATCH;
		return MSS_ERROR_RESP_UNPACK;
	}
	return 0;
}

/**
 *  Fast I/O, support for CEATA devices.
 */
static int mmc_unpack_r4( struct mss_cmd *cmd, struct mmc_response_r4 *r4, struct mmc_card *mmc_card)
{
	u8 *buf = cmd->response;

	mmc_card->errno = MMC_ERROR_NONE;

	r4->rca = PARSE_U16(buf);
	r4->status = buf[3]>>7;
	r4->reg_addr = buf[3] & 0x7f;
	r4->read_reg_contents = buf[4];
	
	return 0 ;
}

/**
 *  Interrupt request. not supported temporarily.
 */
static int mmc_unpack_r5(struct mss_cmd *cmd, struct mmc_response_r5 *r5, struct mmc_card *mmc_card)
{
	mmc_card->errno = MMC_ERROR_NONE;
	return 0 ;
}

static int mmc_unpack_cid(struct mss_cmd *cmd, struct mmc_cid *cid, struct mmc_card *mmc_card)
{
	u8 *buf = cmd->response;
	int i;

	mmc_card->errno = MMC_ERROR_NONE;

	cid->mid = buf[1];
	cid->oid = PARSE_U16(buf + 2);
	for (i = 0; i < 5; i++)
		cid->pnm[i] = buf[4 + i];
	cid->pnm[5] = 0;
	cid->prv = buf[9];
	cid->psn = PARSE_U32(buf + 10);
	cid->mdt = PARSE_U16(buf + 14);
#if 0	
	DEBUG(" mid=%d oid=%d pnm=%s prv=%d.%d psn=%08x mdt=%d/%d\n",
	      cid->mid, cid->oid, cid->pnm, 
	      (cid->prv>>4), (cid->prv&0xf), 
	      cid->psn, cid->mdt&&0xf, ((cid->mdt>>4)&0xff)+2000);
#endif
	if (buf[0] != 0x3f)  {
	       mmc_card->errno = MMC_ERROR_HEADER_MISMATCH;
	       return MSS_ERROR_RESP_UNPACK;
	}
      	return 0;
}

static int mmc_unpack_csd(struct mss_cmd *cmd, struct mmc_csd *csd, struct mmc_card *mmc_card)
{
	u8 *buf = cmd->response;
	
	mmc_card->errno = MMC_ERROR_NONE;

	csd->csd_structure      = (buf[1] & 0xc0) >> 6;
	csd->spec_vers          = (buf[1] & 0x3c) >> 2;
	csd->taac               = buf[2];
	csd->nsac               = buf[3];
	csd->tran_speed         = buf[4];
	csd->ccc                = (((u16)buf[5]) << 4) | ((buf[6] & 0xf0) >> 4);
	csd->read_bl_len        = buf[6] & 0x0f;
	csd->read_bl_partial    = (buf[7] & 0x80) ? 1 : 0;
	csd->write_blk_misalign = (buf[7] & 0x40) ? 1 : 0;
	csd->read_blk_misalign  = (buf[7] & 0x20) ? 1 : 0;
	csd->dsr_imp            = (buf[7] & 0x10) ? 1 : 0;
	csd->c_size             = ((((u16)buf[7]) & 0x03) << 10) | (((u16)buf[8]) << 2) | (((u16)buf[9]) & 0xc0) >> 6;
	csd->vdd_r_curr_min     = (buf[9] & 0x38) >> 3;
	csd->vdd_r_curr_max     = buf[9] & 0x07;
	csd->vdd_w_curr_min     = (buf[10] & 0xe0) >> 5;
	csd->vdd_w_curr_max     = (buf[10] & 0x1c) >> 2;
	csd->c_size_mult        = ((buf[10] & 0x03) << 1) | ((buf[11] & 0x80) >> 7);
	switch (csd->csd_structure ) {
		case CSD_STRUCT_1_0:
		case CSD_STRUCT_1_1:
			csd->erase.v22.sector_size    = (buf[11] & 0x7c) >> 2;
			csd->erase.v22.erase_grp_size = ((buf[11] & 0x03) << 3)
			       	| ((buf[12] & 0xe0) >> 5);
		break;
	case CSD_STRUCT_1_2:
	default:
		csd->erase.v31.erase_grp_size = (buf[11] & 0x7c) >> 2;
		csd->erase.v31.erase_grp_mult = ((buf[11] & 0x03) << 3) | ((buf[12] & 0xe0) >> 5);
		break;
	}
	csd->wp_grp_size        = buf[12] & 0x1f;
	csd->wp_grp_enable      = (buf[13] & 0x80) ? 1 : 0;
	csd->default_ecc        = (buf[13] & 0x60) >> 5;
	csd->r2w_factor         = (buf[13] & 0x1c) >> 2;
	csd->write_bl_len       = ((buf[13] & 0x03) << 2) | ((buf[14] & 0xc0) >> 6);
	csd->write_bl_partial   = (buf[14] & 0x20) ? 1 : 0;
	csd->file_format_grp    = (buf[15] & 0x80) ? 1 : 0;
	csd->copy               = (buf[15] & 0x40) ? 1 : 0;
	csd->perm_write_protect = (buf[15] & 0x20) ? 1 : 0;
	csd->tmp_write_protect  = (buf[15] & 0x10) ? 1 : 0;
	csd->file_format        = (buf[15] & 0x0c) >> 2;
	csd->ecc                = buf[15] & 0x03;
#if 0 
	printk("  csd_structure=%d  spec_vers=%d  taac=%02x  nsac=%02x  tran_speed=%02x\n"
	      "  ccc=%04x  read_bl_len=%d  read_bl_partial=%d  write_blk_misalign=%d\n"
	      "  read_blk_misalign=%d  dsr_imp=%d  c_size=%d  vdd_r_curr_min=%d\n"
	      "  vdd_r_curr_max=%d  vdd_w_curr_min=%d  vdd_w_curr_max=%d  c_size_mult=%d\n"
	      "  wp_grp_size=%d  wp_grp_enable=%d  default_ecc=%d  r2w_factor=%d\n"
	      "  write_bl_len=%d  write_bl_partial=%d  file_format_grp=%d  copy=%d\n"
	      "  perm_write_protect=%d  tmp_write_protect=%d  file_format=%d  ecc=%d\n",
	      csd->csd_structure, csd->spec_vers, 
	      csd->taac, csd->nsac, csd->tran_speed,
	      csd->ccc, csd->read_bl_len, 
	      csd->read_bl_partial, csd->write_blk_misalign,
	      csd->read_blk_misalign, csd->dsr_imp, 
	      csd->c_size, csd->vdd_r_curr_min,
	      csd->vdd_r_curr_max, csd->vdd_w_curr_min, 
	      csd->vdd_w_curr_max, csd->c_size_mult,
	      csd->wp_grp_size, csd->wp_grp_enable,
	      csd->default_ecc, csd->r2w_factor, 
	      csd->write_bl_len, csd->write_bl_partial,
	      csd->file_format_grp, csd->copy, 
	      csd->perm_write_protect, csd->tmp_write_protect,
	      csd->file_format, csd->ecc);
	switch ( csd->csd_structure ) {
	case CSD_STRUCT_1_0:
	case CSD_STRUCT_1_1:
		printk(" V22 sector_size=%d erase_grp_size=%d\n", 
		      csd->erase.v22.sector_size, 
		      csd->erase.v22.erase_grp_size);
		break;
	case CSD_STRUCT_1_2:
	default:
		printk(" V31 erase_grp_size=%d erase_grp_mult=%d\n", 
		      csd->erase.v31.erase_grp_size,
		      csd->erase.v31.erase_grp_mult);
		break;
		
	}
#endif
	if (buf[0] != 0x3f)  {
		mmc_card->errno = MMC_ERROR_HEADER_MISMATCH;
		return MSS_ERROR_RESP_UNPACK;
	}

	return MSS_ERROR_NONE;
}

static int mmc_unpack_ext_csd(struct mss_cmd *cmd, struct mmc_ext_csd *ext_csd, struct mmc_card *mmc_card)
{
	char *buf = mmc_card->buf;

	mmc_card->errno = MMC_ERROR_NONE;
	ext_csd->s_cmd_set = buf[504];
	ext_csd->sec_count = ((u32)buf[212]) << 24 | ((u32)buf[213]) << 16 | ((u32)buf[214]) << 8 | ((u32)buf[215]) ;
	ext_csd->min_perf_w_8_52 = buf[210];
	ext_csd->min_perf_r_8_52 = buf[209];
	ext_csd->min_perf_w_26_4_52 = buf[208];
	ext_csd->min_perf_r_26_4_52 = buf[207];
	ext_csd->min_perf_w_4_26 = buf[206];
	ext_csd->min_perf_r_4_26 = buf[205];
	ext_csd->pwr_cl_26_360 = buf[203];
	ext_csd->pwr_cl_52_360 = buf[202];
	ext_csd->pwr_cl_26_195 = buf[201];
	ext_csd->pwr_cl_52_195 = buf[200];
	ext_csd->card_type = buf[196];
	ext_csd->csd_structure = buf[194];
	ext_csd->ext_csd_rev = buf[192];
	ext_csd->cmd_set = buf[191];
	ext_csd->cmd_set_rev = buf[189];
	ext_csd->power_class = buf[187];
	ext_csd->hs_timing = buf[185];
	ext_csd->erased_mem_cont = buf[183];

#if 0
	DEBUG("s_cmd_set:%d, sec_count:%d, min_perf_w_8_52:%d,\n"
	"min_perf_r_8_52:%d, min_perf_w_26_4_52:%d, min_perf_r_26_4_52:%d\n"
	"min_perf_w_4_26:%d, min_perf_r_4_26:%d, pwr_cl_26_360:%d, "
	"pwr_cl_52_360:%d, pwr_cl_26_195:%d, pwr_cl_52_195:%d, card_type:%d\n"
	"csd_structure:%d, ext_csd_rev:%d, cmd_set:%d, cmd_set_rev:%d\n"
	"power_class:%d, hs_timing:%d, erased_mem_cont:%d\n", ext_csd->s_cmd_set,ext_csd->sec_count,ext_csd->min_perf_w_8_52, ext_csd->min_perf_r_8_52,ext_csd->min_perf_w_26_4_52,ext_csd->min_perf_r_26_4_52, ext_csd->min_perf_w_4_26, ext_csd->min_perf_r_4_26, ext_csd->pwr_cl_26_360, ext_csd->pwr_cl_52_360, ext_csd->pwr_cl_26_195, ext_csd->pwr_cl_52_195, ext_csd->card_type, ext_csd->csd_structure, ext_csd->ext_csd_rev, ext_csd->cmd_set, ext_csd->cmd_set_rev, ext_csd->power_class, ext_csd->hs_timing, ext_csd->erased_mem_cont); 
#endif
	return 0;
}

/**
 *  The blocks requested by the kernel may or may not match what we can do.  
 *  Unfortunately, filesystems play fast and loose with block sizes, so we're 
 * stuck with this .
 */
static void mmc_fix_request_block_len(struct mss_card *card, int action, struct mss_rw_arg *arg)
{
	u16 block_len = 0;
	struct mmc_card *mmc_card = card->prot_card;
	struct mss_host *host = card->slot->host;

	switch(action) {
		case MSS_DATA_READ:
			block_len = 1 << mmc_card->csd.read_bl_len;
			break;
		case MSS_DATA_WRITE:
			block_len = 1 << mmc_card->csd.write_bl_len;
			break;
		default:
			return;
	}
	if (host->high_capacity && mmc_card->access_mode == MMC_ACCESS_MODE_SECTOR)
		block_len = 512;	
	if (block_len < arg->block_len) {
		int scale = arg->block_len / block_len;
		arg->block_len	= block_len;
		arg->block *= scale;
		arg->nob *= scale;
	}
}


/*****************************************************************************
 *
 *   protocol entry functions
 *
 ****************************************************************************/

static u32 mmc_make_cmd6_arg(u8 access, u8 index, u8 value, u8 cmdset)
{
        u32 ret;
        ret = (access << 24) | (index << 16) | (value << 8) | cmdset;
        return ret;
}

static int mmc_simple_ll_req(struct mss_host *host, struct mmc_card *mmc_card, u32 opcode, u32 arg, u32 rtype, u32 flags)
{
	memset(&mmc_card->llreq, 0x0, sizeof(struct mss_ll_request));
	memset(&mmc_card->cmd, 0x0, sizeof(struct mss_cmd));

	mmc_card->cmd.opcode = opcode;
	mmc_card->cmd.arg = arg;
	mmc_card->cmd.rtype = rtype;
	mmc_card->cmd.flags = flags;

	mmc_card->llreq.cmd = &mmc_card->cmd;

	return mss_send_ll_req(host, &mmc_card->llreq);
}	

static int mmc_get_status(struct mss_card *card, u32 *status)
{
	struct mmc_response_r1 r1;
	struct mmc_card *mmc_card = card->prot_card;
	struct mss_host *host = card->slot->host;
	int clock, ret, retries = 4;

	clock = mmc_tran_speed(mmc_card->csd.tran_speed);
	mss_set_clock(card->slot->host, clock);
	while (retries--) {
		dbg5("rety");
		ret = mmc_simple_ll_req(host, mmc_card, MMC_SEND_STATUS, 
			MSS_SLOT_RCA(card->slot) << 16, MSS_RESPONSE_R1, 0);
		dbg5("retry ret :%d", ret);
		if (ret && !retries)
			return ret;
		else if (!ret) {
			ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
			if (ret) {
				if (mmc_card->errno == MMC_ERROR_STATE_MISMATCH) {
					mmc_card->state = R1_CURRENT_STATE(r1.status);
					mmc_card->errno = MMC_ERROR_NONE;
				}
				else
					return ret;
			}
			else 
				break;
		}

		clock = host->ios.clock;
		clock = clock >> 1;
		if (clock < MMC_CARD_CLOCK_SLOW || retries == 1)
			clock = MMC_CARD_CLOCK_SLOW;
		mss_set_clock(host, clock);
	}

	*status = r1.status;
	
	return MSS_ERROR_NONE; 
}

static int mmc_recognize_card(struct mss_card *card)
{
	struct mmc_response_r3 r3;
	struct mmc_card *mmc_card = card->prot_card;
	struct mss_ios ios;
	struct mss_host *host = card->slot->host;
	int ret;
	u32 ocr = host->vdd;

	mmc_card->state = CARD_STATE_IDLE;
	card->bus_width = MSS_BUSWIDTH_1BIT;
	card->card_type = MSS_UNKNOWN_CARD;
	
	memcpy(&ios, &host->ios, sizeof(struct mss_ios)); 
	ios.bus_mode = MSS_BUSMODE_OPENDRAIN;
	ios.clock = host->clk_min;
	ios.bus_width = MSS_BUSWIDTH_1BIT;
	host->ops->set_ios(host, &ios);

	ret = mmc_simple_ll_req(host, mmc_card, MMC_GO_IDLE_STATE, 0, MSS_RESPONSE_NONE, MSS_CMD_INIT); 
	dbg2(card->slot, card->prot_driver, "Sending GO_IDLE cmd, ret:%d\n", ret);
	if (ret)
		return ret;
	if (host->high_capacity)
		ocr |= MMC_ACCESS_MODE_SECTOR;
	ret = mmc_simple_ll_req(host, mmc_card, MMC_SEND_OP_COND, ocr, MSS_RESPONSE_R3, 0); 
	dbg2(card->slot, card->prot_driver, "Sending SEND_OP_COND cmd, arg:0x%x\n, ret:%d", ocr, ret);
	if (ret)
		return ret;
	
	ret = mmc_unpack_r3(&mmc_card->cmd, &r3, mmc_card); 
	dbg2(card->slot, card->prot_driver, "unapck ret %d, SEND_OP_COND ocr:0x%x", ret, r3.ocr);
	if (!ret) {
		if (r3.ocr & host->vdd) {
			card->card_type = MSS_MMC_CARD;
		} else {
			printk(KERN_WARNING "uncompatible card\n");
			card->card_type = MSS_UNCOMPATIBLE_CARD;	
		}
		return ret;
	}
	card->card_type = MSS_UNKNOWN_CARD;

	return MSS_ERROR_NONE;
}

#if 0
#ifdef CONFIG_CEATA

/*
* check whether current card is CEATA device
*/

static int ceata_discovery(struct mss_card_device *dev)
{
	struct mmc_response_r1 r1;
	struct mmc_response_r4 r4;
	int retval;
	int status;
	u8 *buffer;
	int i;

	u8 ext_csd[512];
	u8 task_file[32];
	u8 ident_device_data[512];
	
	int retries =10;
	buffer = NULL;
	
	retval = mmc_get_status(dev, &status);
	if(retval) {
		DEBUG("can not get card status");
		retval = MSS_READ_WRITE_ERROR;
		goto ceata_discovery_error;
	}

	if(status & CARD_STATUS_LOCKED) {
		printk("card is locked\n");
		retval = MSS_LOCKED;
		goto ceata_discovery_error;
	}
	
	if (dev->state == CARD_STATE_STBY) {
		mmc_fix_request_block_size(dev);
		mss_simple_cmd( dev, MMC_SELECT_CARD, (dev->slot->rca) << 16, RESPONSE_R1B, buffer );
		if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
			goto ceata_discovery_error;
	}
	dev->state = CARD_STATE_TRAN;

	// issusing CMD39 to check the CEATA signture

	// check address of LBA Middle register(DH)  - CE
	mss_simple_cmd(dev, MMC_FAST_IO, ((dev->slot->rca) << 16) + (CEATA_LBA_MID<<8), RESPONSE_R4, buffer );

	retval  = mmc_unpack_r4(dev->io_request, &r4, dev->state);
	if(retval)
		goto ceata_discovery_error;

	if(!(r4.read_reg_contents == 0xCE)){
		dbg("CMD39 for CEATA_LBA_MID data is :%d", r4.read_reg_contents);
		goto ceata_discovery_error;
	}


	// check address of LBA High register(Ch) - AA
	mss_simple_cmd(dev, MMC_FAST_IO, ((dev->slot->rca) <<16) + (CEATA_LBA_HIGH<<8), RESPONSE_R4, buffer );

	retval  = mmc_unpack_r4(dev->io_request, &r4, dev->state);
	if(retval)
		goto ceata_discovery_error;

	if(!(r4.read_reg_contents == 0xAA)){
		dbg("CMD39 for CEATA_LBA_HIGH data is :%d", r4.read_reg_contents);
		goto ceata_discovery_error;
	}

	dbg("1.CEATA signature verified");
	
	// repeat CMD39 to check the device state
	do{
		mss_simple_cmd(dev, MMC_FAST_IO, ((dev->slot->rca) <<16) + (CEATA_COMMAND_STATUS<<8), RESPONSE_R4, buffer );
		retval  = mmc_unpack_r4(dev->io_request, &r4, dev->state);
		if(retval)
			goto ceata_discovery_error;

		status = r4.read_reg_contents;

		mdelay(200);

	}while((status!= 0x40) &(retries -- >0));

	dbg("2.CEATA device is ready for read/write");


	// get EXT_CSD register data
	buffer = ext_csd;
	dev->cmd_flag = MSS_READ;
	mss_send_cmd(dev, MMC_SEND_EXT_CSD, 0, 1, 512, RESPONSE_R1, buffer);
	if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
		goto ceata_discovery_error;
//	for(i=0; i< 512; i++)
//		printk("%d, %x\n",i, ext_csd[i]); 
	dbg("3.CEATA receive the EXT_CSD register");


	// todo: check whether the current one support 4-bit mode
	
	// use CMD6 to switch to 4-bit mode
	buffer = NULL;
	mss_simple_cmd(dev, MMC_SWITCH, 0x3b70100 , RESPONSE_R1B, buffer );
	if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
		goto ceata_discovery_error;

	dev->bus_mode = MSS_4_BIT;
	dbg("4.CEATA success switch to 4-bit mode");


	// read CEATA identify data

	//step1: prepare the task file, and issue CMD60
	for(i=0; i<32; i++){
		task_file[i] = 0;
	}
#ifndef CEATA_CSS_MODE	
	task_file[CEATA_CONTROL] = 0x02;  // not using CSS
#endif

	task_file[CEATA_COMMAND_STATUS] = CEATA_IDENTIFY_DATA;

	buffer = task_file;
	dev->cmd_flag = MSS_WRITE;
	mss_send_cmd(dev, CEATA_RW_MULTIPLE_REGISTER, 0x80000010, 1, 16, RESPONSE_R1B, buffer);
	if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
		goto ceata_discovery_error;

	// todo: check the status of the CEATA device
	// step2: use CMD39 to check the device status
	// repeat CMD39 to check the device state
	buffer = NULL;
	do{
		mss_simple_cmd(dev, MMC_FAST_IO, ((dev->slot->rca) <<16) + (CEATA_COMMAND_STATUS<<8), RESPONSE_R4, buffer );
		retval  = mmc_unpack_r4(dev->io_request, &r4, dev->state);
		if(retval)
			goto ceata_discovery_error;

		status = r4.read_reg_contents;

		dbg("CEATA status:%x", r4.read_reg_contents);
		mdelay(200);

	}while(((status && 0x40)!= 1) &(retries -- >0));

	// step3: get the CEATA identify data back from address 0
	buffer = ident_device_data;
	dev->cmd_flag = MSS_READ;
	for(i=0; i<512; i++)
		ident_device_data[i] = 0;

	mss_send_cmd(dev, CEATA_RW_MULTIPLE_BLOCK, 0x1, 1, 512, RESPONSE_R1B, buffer);
	if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
		goto ceata_discovery_error;

		
#if 0
	for(i=0; i< 512; i++)
		printk("%d, %x\n",i, ident_device_data[i]); 
#endif
	dbg("5.CEATA receive the identify device data");

	return MSS_SUCCESS;

ceata_discovery_error:
	debug("5. in ceata_discovery, error occured. RETVAL: %d, SLOT->STATE: %d\n", retval, dev->state);
//	mss_finish_io_request(dev, 0);
	if(retval == 0)
		retval = MSS_ERROR;
	return retval;
	
}

#endif
#endif

static int mmc_card_init(struct mss_card *card)
{
	struct mmc_response_r3 r3;
	struct mmc_response_r1 r1;
	struct mmc_cid cid;
	struct mmc_card *mmc_card = card->prot_card;
	struct mss_ios ios;
	struct mss_host *host;
	int ret;
	u8 *g_buffer;
	u32 ocr, arg, bus_width;
	struct scatterlist sg;
       
	host = card->slot->host;
	ocr = host->vdd;

	mmc_card->state = CARD_STATE_IDLE;
	card->bus_width = MSS_BUSWIDTH_1BIT;
	
	memcpy(&ios, &host->ios, sizeof(struct mss_ios)); 
	ios.bus_mode = MSS_BUSMODE_OPENDRAIN;
	ios.clock = host->clk_min;
	host->ops->set_ios(host, &ios);

	ret = mmc_simple_ll_req(host, mmc_card, MMC_GO_IDLE_STATE, 0, MSS_RESPONSE_NONE, MSS_CMD_INIT); 
	dbg2(card->slot, card->prot_driver, "Sending GO_IDLE cmd, ret:%d\n", ret);
	if (ret)
		return ret;

	if (host->high_capacity)
		ocr |= MMC_ACCESS_MODE_SECTOR;
	ret = mmc_simple_ll_req(host, mmc_card, MMC_SEND_OP_COND, ocr, MSS_RESPONSE_R3, 0); 
	dbg2(card->slot, card->prot_driver, "Sending SEND_OP_COND cmd, arg:0x%x\n, ret:%d", ocr, ret);
	if (ret)
		return ret;
	
	ret = mmc_unpack_r3(&mmc_card->cmd, &r3, mmc_card);
	dbg2(card->slot, card->prot_driver, "unapck ret %d, SEND_OP_COND ocr:0x%x", ret, r3.ocr);
	if (ret) {
		printk(KERN_ERR "failed SEND_OP_COND error=%d (%s) \n", ret, mmc_result_to_string(ret));
		return ret;
	}

	while (!(r3.ocr & MMC_CARD_BUSY)) {
		mdelay(20);
		dbg2(card->slot, card->prot_driver, "card is busy after SNED_OP_COND");
		ret = mmc_simple_ll_req(host, mmc_card, MMC_SEND_OP_COND, ocr, MSS_RESPONSE_R3, 0); 
		dbg2(card->slot, card->prot_driver, "Sending SEND_OP_COND cmd, arg:0x%x\n, ret:%d", ocr, ret);
		if (ret)
			return ret;
		ret = mmc_unpack_r3(&mmc_card->cmd, &r3, mmc_card);
		dbg2(card->slot, card->prot_driver, "unapck ret %d, SEND_OP_COND ocr:0x%x", ret, r3.ocr);
		if (ret) {
			printk(KERN_ERR ": failed SEND_OP_COND error=%d (%s) \n", retval, mmc_result_to_string(retval) );
			return ret;
		}
	}

	mmc_card->vdd = r3.ocr & MMC_VDD_MASK;
	mmc_card->access_mode = r3.ocr & MMC_ACCESS_MODE_MASK;
	mmc_card->state = CARD_STATE_READY;
	
	ret = mmc_simple_ll_req(host, mmc_card, MMC_ALL_SEND_CID, 0, MSS_RESPONSE_R2_CID, 0); 
	dbg2(card->slot, card->prot_driver, "Sending MMC_ALL_SEND_CID cmd, arg:0x%x\n, ret:%d", 0, ret);
	if (ret)
		return ret;

	ret = mmc_unpack_cid(&mmc_card->cmd, &cid, mmc_card);
	if (ret)
		return ret;
	
	/* Not first init */
	if (mmc_card->cid.mid != 0) {
		/* Card is changed */
		if (memcmp(&cid, &mmc_card->cid, sizeof(cid)))
			return MSS_ERROR_MISMATCH_CARD;
	}
	else 
		memcpy(&mmc_card->cid, &cid, sizeof(cid));
	
	mmc_card->state = CARD_STATE_IDENT;

	mss_set_busmode(host, MSS_BUSMODE_PUSHPULL);

	ret = mmc_simple_ll_req(host, mmc_card, MMC_SET_RELATIVE_ADDR, 
			MSS_SLOT_RCA(card->slot) << 16, MSS_RESPONSE_R1, 0); 
	dbg2(card->slot, card->prot_driver, "Sending SET_REL_ADDR cmd, arg:0x%x\n, ret:%d", MSS_SLOT_RCA(card->slot) << 16, ret);
	if (ret)
		return ret;

	//mss_simple_cmd(dev, MMC_SET_RELATIVE_ADDR, (dev->slot->rca) << 16, RESPONSE_R1, buffer);
    	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	if (ret) {
		printk(KERN_INFO "MMC: unable to SET_RELATIVE_ADDR error=%d (%s)\n", retval, mmc_result_to_string(retval) );
		return ret;
	}
  	mmc_card->state = CARD_STATE_STBY;

	ret = mmc_simple_ll_req(host, mmc_card, MMC_SEND_CSD, 
			MSS_SLOT_RCA(card->slot) << 16, MSS_RESPONSE_R2_CSD, 0);
	dbg2(card->slot, card->prot_driver, "Sending MMC_SEND_CSD cmd, arg:0x%x\n, ret:%d", MSS_SLOT_RCA(card->slot) << 16, ret);
	if (ret)
		return ret;
	ret = mmc_unpack_csd(&mmc_card->cmd, &mmc_card->csd, mmc_card);
	if (ret) {
		printk(KERN_INFO "MMC: unable to SEND_CSD error=%d (%s)\n", retval, mmc_result_to_string(retval) );
		return ret;
	}
	
	/*
	 * if it is MMC4.x-compatible card and how many bits are working.
	 */
	if (mmc_card->csd.spec_vers != CSD_SPEC_VERS_4 
			|| host->mmc_spec != MSS_MMC_SPEC_40_42) 
		goto exit;

	g_buffer = mmc_card->buf;//kmalloc(512, GFP_KERNEL);
	/*if (!g_buffer)
		return -ENOMEM;		*/
	ret = mmc_simple_ll_req(host, mmc_card, MMC_SELECT_CARD, 
			MSS_SLOT_RCA(card->slot) << 16, MSS_RESPONSE_R1B, 0);
	if (ret)
		return ret;
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	if (ret)
		return ret;
	mmc_card->state = CARD_STATE_TRAN;

	/*
	 * set 1-bus mode in init. arg:access=0b11, arg:index=0xb7, arg:value=0
	 * or CMD8 will not be responded with 1-bit/controller and 4-bit/card
	 * dev->bus_mode should be MSS_1_BIT before next command
	 * buffer = NULL;
	 */
	arg = mmc_make_cmd6_arg(0x3, 0xb7, 0, 0);
	ret = mmc_simple_ll_req(host, mmc_card, MMC_SWITCH, arg , 
			MSS_RESPONSE_R1B, 0);
	if (ret)
		return ret;
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	if (ret)
		return ret;
	
	card->bus_width = MSS_BUSWIDTH_1BIT;
	mss_set_buswidth(host, MSS_BUSWIDTH_1BIT);

	memset(&mmc_card->llreq, 0x0, sizeof(struct mss_ll_request));
	memset(&mmc_card->cmd, 0x0, sizeof(struct mss_cmd));
	mmc_card->cmd.opcode = MMC_SEND_EXT_CSD;
	mmc_card->cmd.arg = 0;
	mmc_card->cmd.rtype = MSS_RESPONSE_R1;
	mmc_card->cmd.flags = 0;

	mmc_card->data.nob = 1;
	/* 512B is supported by all the card */
	mmc_card->data.block_len = 512;
	mmc_card->data.flags = MSS_DATA_READ;
	mmc_card->data.sg_len = 1;
	mmc_card->data.sg = &sg;
	sg.page = virt_to_page(g_buffer);
	sg.offset = offset_in_page(g_buffer);
	sg.length = 512;

	mmc_card->llreq.cmd = &mmc_card->cmd;
	mmc_card->llreq.data = &mmc_card->data;
	ret = mss_send_ll_req(host, &mmc_card->llreq);
	if (ret)
		return ret;
	ret = mmc_unpack_ext_csd(&mmc_card->cmd, &mmc_card->ext_csd, mmc_card);
	if (ret)
		return ret;


	/*
	 *  use CMD19/CMD14 (BUSTEST_W/BUSTEST_R) to test supported bus mode.
	 */
	memset(g_buffer, 0xFF, 512);
	bus_width = host->bus_width;
	if (bus_width == MSS_BUSWIDTH_1BIT)
		goto exit;
	else if (bus_width == MSS_BUSWIDTH_4BIT) {
		card->bus_width = MSS_BUSWIDTH_4BIT;
		g_buffer[0] = 0xa5; /* 0b10100101 */
		g_buffer[1] = 0xFF; /* 0b11111111 */
	}
	else if (bus_width == MSS_BUSWIDTH_8BIT) {
		card->bus_width = MSS_BUSWIDTH_8BIT;
		g_buffer[0] = 0xaa; /* 0b10101010 */
		g_buffer[1] = 0x55; /* 0b01010101 */
	}
	else
		goto exit;
	
	mss_set_buswidth(host, bus_width);
	
	memset(&mmc_card->llreq, 0x0, sizeof(struct mss_ll_request));
	memset(&mmc_card->cmd, 0x0, sizeof(struct mss_cmd));
	mmc_card->cmd.opcode = MMC_BUSTEST_W;
	mmc_card->cmd.arg = 0;
	mmc_card->cmd.rtype = MSS_RESPONSE_R1;
	mmc_card->cmd.flags = 0;

	mmc_card->data.nob = 1;
	mmc_card->data.block_len = 512;
	mmc_card->data.flags = MSS_DATA_WRITE;
	mmc_card->data.sg_len = 1;
	mmc_card->data.sg = &sg;
	sg.page = virt_to_page(g_buffer);
	sg.offset = offset_in_page(g_buffer);
	sg.length = 512;

	mmc_card->llreq.cmd = &mmc_card->cmd;
	mmc_card->llreq.data = &mmc_card->data;
	ret = mss_send_ll_req(host, &mmc_card->llreq);
	/* FIXME: PXA MMC */
	/*if (ret)
		return ret;*/
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	if (ret)
		return ret;

	mmc_card->state = CARD_STATE_BTST;
	memset(g_buffer, 0xFF, 512);

	memset(&mmc_card->llreq, 0x0, sizeof(struct mss_ll_request));
	memset(&mmc_card->cmd, 0x0, sizeof(struct mss_cmd));
	mmc_card->cmd.opcode = MMC_BUSTEST_R;
	mmc_card->cmd.arg = 0;
	mmc_card->cmd.rtype = MSS_RESPONSE_R1;
	mmc_card->cmd.flags = 0;

	mmc_card->data.nob = 1;
	mmc_card->data.block_len = 512;
	mmc_card->data.flags = MSS_DATA_READ;
	mmc_card->data.sg_len = 1;
	mmc_card->data.sg = &sg;
	sg.page = virt_to_page(g_buffer);
	sg.offset = offset_in_page(g_buffer);
	sg.length = 512;

	mmc_card->llreq.cmd = &mmc_card->cmd;
	mmc_card->llreq.data = &mmc_card->data;
	ret = mss_send_ll_req(host, &mmc_card->llreq);
	/* FIXME: PXA MMC */
	/*if (ret)
		return ret;*/
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	if (ret)
		return ret;

	if ((g_buffer[0] == 0x55/*0b01010101*/) && 
		(g_buffer[1] == 0xaa/*0b10101010*/)) {
		mmc_card->bus_width = MSS_BUSWIDTH_8BIT;
	}
	else if (g_buffer[0] == 0x5a /*0b01011010*/) {
		mmc_card->bus_width = MSS_BUSWIDTH_4BIT;
	}
	else { 
		mmc_card->bus_width = MSS_BUSWIDTH_1BIT;
	}

	mmc_card->state = CARD_STATE_TRAN;
	
	arg = mmc_make_cmd6_arg(0x3, 0xb7, mmc_card->bus_width, 0);
	ret = mmc_simple_ll_req(host, mmc_card, MMC_SWITCH, arg , 
			MSS_RESPONSE_R1B, 0);
	if (ret)
		return ret;
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	if (ret)
		return ret;
	
	card->bus_width = mmc_card->bus_width;
	/** 
	 * use CMD6 to set high speed mode. arg:access=0b11, arg:index=0xb9, 
	 * arg:value=1 according to card/controller high speed timing  
	 * high speed mode for MMC-4.x compatible card.
	 */
	if (host->high_speed == MSS_HIGH_SPEED) {
		mmc_make_cmd6_arg(0x3, 0xb9, 1, 0);
		ret = mmc_simple_ll_req(host, mmc_card, MMC_SWITCH, arg , 
				MSS_RESPONSE_R1B, 0);
		if (ret)
			return ret;
		ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
		if (ret)
			return ret;
		memcpy(&ios, &host->ios, sizeof(struct mss_ios)); 
		ios.high_speed = MSS_HIGH_SPEED;
		ios.clock = MMC_CARD_CLOCK_FAST;
		host->ops->set_ios(host, &ios);	
		if (host->ios.high_speed != MSS_HIGH_SPEED) {
			u32 clock = mmc_tran_speed(mmc_card->csd.tran_speed);
			mss_set_clock(host, clock);	
		}
	}
	else {
		/* change to the highest speed in normal mode */
		u32 clock = mmc_tran_speed(mmc_card->csd.tran_speed);
		mss_set_clock(host, clock);
	}
		
	/** 
	 * use CMD6 to set power class. arg:access=0b11, arg:index=0xbb, 
	 * arg:value=card power class code according to card/controller supported 
	 * power class. We read value from PWL_CL_ff_vvv and set it to POWER_CLASS
	 * according to supported voltage and clock rate.
	 */

	/* Deselect the card */
	ret = mmc_simple_ll_req(host, mmc_card, MMC_SELECT_CARD, 0, MSS_RESPONSE_R1B, 0);
	if (ret)
		return ret;
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	if (ret)
		return ret;
	mmc_card->state = CARD_STATE_STBY;

exit:
	return MSS_ERROR_NONE;
}
#if 0
#ifdef CONFIG_CEATA
static enum mss_result ceata_read_write_entry(struct mss_card_device *dev)
{
	int status = 0;
	struct mss_io_request *t = dev->io_request;
	int nr_sectors = t->nr_sectors;
	int retval = 0;
	int i;
	u32 arg;
	struct mmc_response_r1 r1;
	struct mmc_response_r4 r4;

//	int retries = 1;
	u8 *buffer;
	int cmd_backup;
	
	u8 task_file[32];

	dbg("card: 0x%p, io_request: 0x%p", dev, dev->io_request);
	buffer = NULL;
	if (dev->card_type != CE_ATA)
		return MSS_WRONG_CARD;

	retval = mmc_get_status(dev, &status);
	if(retval) {
		DEBUG("can not get card status");
		retval = MSS_READ_WRITE_ERROR;
		goto read_write_error;
	}

	if(status & CARD_STATUS_LOCKED) {
		printk("card is locked\n");
		retval = MSS_LOCKED;
		goto read_write_error;
	}
	
	if (dev->state == CARD_STATE_STBY) {
		mmc_fix_request_block_size(dev);
		mss_simple_cmd( dev, MMC_SELECT_CARD, (dev->slot->rca) << 16, RESPONSE_R1B, buffer );
		if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
			goto read_write_error;
	}
	dev->state = CARD_STATE_TRAN;

	dbg("before  MMC_SET_BLOCKLEN");

	mss_simple_cmd(dev, MMC_SET_BLOCKLEN, t->block_len, RESPONSE_R1,buffer);
	if ( (retval = mmc_unpack_r1(dev->io_request, &r1, dev->state)) ) 
		goto read_write_error;

	dbg("after  MMC_SET_BLOCKLEN");

//	while(1);
	
read_write_entry:
//	do {
		cmd_backup = dev->cmd_flag;

		// step1, issue CMD60
 		for(i=0; i<32; i++){
			task_file[i] = 0;
		}
#ifndef CEATA_CSS_MODE	
		task_file[CEATA_CONTROL] = 0x02;  // not using CSS
#endif

		task_file[CEATA_SECTOR_COUNT] = nr_sectors;
		task_file[CEATA_LBA_LOW] = ((t->sector * t->block_len) & 0xFF);
		task_file[CEATA_LBA_MID] = ((t->sector * t->block_len) & 0xFF00)>>8;
		task_file[CEATA_LBA_HIGH] = (t->sector * t->block_len) >>16;
		task_file[CEATA_DEVICE_HEAD] = 0x40;
		task_file[CEATA_COMMAND_STATUS] = (t->action == READ) ? CEATA_READ_DMA_EXT : CEATA_WRITE_DMA_EXT;

		buffer = task_file;
		dev->cmd_flag = MSS_WRITE;
		mss_send_cmd(dev, CEATA_RW_MULTIPLE_REGISTER, 0x80000010, 1, 16, RESPONSE_R1B, buffer);
		if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
			goto read_write_error;

		dbg("1.  CEATA_RW_MULTIPLE_REGISTER complete");

		// step2, issue CMD39 to check CEATA device status -- whether ready for data transfer
		buffer = NULL;
		do{
			mss_simple_cmd(dev, MMC_FAST_IO, ((dev->slot->rca) <<16) + (CEATA_COMMAND_STATUS<<8), RESPONSE_R4, buffer );
			retval  = mmc_unpack_r4(dev->io_request, &r4, dev->state);
			if(retval)
				goto read_write_error;

			status = r4.read_reg_contents;

			dbg("CEATA status:%x", r4.read_reg_contents);
			mdelay(200);
		}while(((status & 0x48)!= 0x48)) ;
//		}while(((status && 0x40)!= 1) &(retries -- >0));

		dbg("2.  first MMC_FAST_IO complete");

		// step3, issue CMD61 for exact read/write opeation
		dev->cmd_flag = (t->action == READ) ? MSS_READ : MSS_WRITE;
		buffer = dev->io_request->buffer;

		arg = 0;
		if(t->action == WRITE)
			arg = 1<<31;
		arg += nr_sectors;
		mss_send_cmd(dev, CEATA_RW_MULTIPLE_BLOCK, arg, nr_sectors,  t->block_len, RESPONSE_R1B, buffer);
		if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
			goto read_write_error;
		// skip the error operation
/*		{
			debug("1. in read_write_entry, error occured. RETVAL: %d, SLOT->STATE: %d, RETRY: %d, CARD RATE: %d\n",	retval, dev->state, retries, mmc_tran_speed(((struct mmc_card_info *)dev->card)->card_csd.tran_speed));
			dev->state = (t->action == READ) ? CARD_STATE_DATA : CARD_STATE_RCV;
			dev->cmd_flag = cmd_backup;
			dbg("1 before send cmd12, buffer: 0x%p", buffer);
			mss_simple_cmd( dev, MMC_STOP_TRANSMISSION, 0, RESPONSE_R1B, buffer  );
			if ( (tmp = mmc_unpack_r1(dev->io_request, &r1, dev->state))) { //just show status, skip for retrying...
				debug("5. in read_write_entry, error occured. RETVAL: %d, SLOT->STATE: %d\n", tmp, dev->state);
		}
*/			

		dbg("3.  CEATA_RW_MULTIPLE_BLOCK complete");
	
		// step4, issue CMD39 again to CEATA status
		buffer = NULL;
		do{
			mss_simple_cmd(dev, MMC_FAST_IO, ((dev->slot->rca) <<16) + (CEATA_COMMAND_STATUS<<8), RESPONSE_R4, buffer );
			retval  = mmc_unpack_r4(dev->io_request, &r4, dev->state);
			if(retval)
				goto read_write_error;

			status = r4.read_reg_contents;

			dbg("CEATA status:%x", r4.read_reg_contents);
			mdelay(200);
		}while(((status & 0x40)!= 0x40)) ;

		dbg("4. second MMC_FAST_IO complete");

		dev->state = CARD_STATE_TRAN;
//		while(1);
//	} while (retval && (retries-- > 0));
//	if (retval) 
//		goto read_write_error;
	mss_finish_io_request(dev, 1);	
	return MSS_SUCCESS;

read_write_error:
	debug("5. in read_write_entry, error occured. RETVAL: %d, SLOT->STATE: %d\n", retval, dev->state);
	mss_finish_io_request(dev, 0);	
	return retval;
}
#endif
#endif

static int mmc_read_write_entry(struct mss_card *card, int action, struct mss_rw_arg *arg, struct mss_rw_result *result)
{
	struct mmc_card *mmc_card = card->prot_card;
	struct mss_host *host;
	struct mss_ios ios;
	struct mmc_response_r1 r1;
	int ret, retries = 4, clock;
	u32 status = 0;
	int access_mode_sector = 0;

	host = card->slot->host;

	ret = mmc_get_status(card, &status);
	if (ret)
		return ret;

	if (status & R1_CARD_IS_LOCKED)
		return MSS_ERROR_LOCKED;

	if (action == MSS_WRITE_MEM && host->ops->is_slot_wp && 
			host->ops->is_slot_wp(card->slot))
		return MSS_ERROR_WP;
	
	if (mmc_card->state == CARD_STATE_STBY) {
		ret = mmc_simple_ll_req(host, mmc_card, MMC_SELECT_CARD, 
				MSS_SLOT_RCA(card->slot) << 16, MSS_RESPONSE_R1, 0);
		if (ret)
			return ret;
		dbg5("select card cmd return ret:%d\n", ret);
		ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
		if (ret)
			return ret;
	}
	mmc_card->state = CARD_STATE_TRAN;

	mmc_fix_request_block_len(card, action, arg);
	access_mode_sector = (mmc_card->access_mode == MMC_ACCESS_MODE_SECTOR)
	       	&& host->high_capacity;	
	memcpy(&ios, &host->ios, sizeof(struct mss_ios)); 
	if (access_mode_sector) {
		ios.access_mode = MSS_ACCESS_MODE_SECTOR;
	}
	else {
		ret = mmc_simple_ll_req(host, mmc_card, MMC_SET_BLOCKLEN, arg->block_len, MSS_RESPONSE_R1, 0);
		ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card); 
		if (ret)
			return ret;
	}
	if (mmc_card->csd.spec_vers == CSD_SPEC_VERS_4 && host->high_speed) {
		ios.high_speed = MSS_HIGH_SPEED;
		ios.clock = MMC_CARD_CLOCK_FAST;
	}
	else {
		clock = mmc_tran_speed(mmc_card->csd.tran_speed);
	}
	host->ops->set_ios(host, &ios);
read_write_entry:
	memset(&mmc_card->llreq, 0x0, sizeof(struct mss_ll_request));
	memset(&mmc_card->cmd, 0x0, sizeof(struct mss_cmd));
	memset(&mmc_card->data, 0x0, sizeof(struct mss_data));
	if (access_mode_sector)
		mmc_card->cmd.arg = arg->block;
	else
		mmc_card->cmd.arg = arg->block * arg->block_len;
	mmc_card->cmd.rtype = MSS_RESPONSE_R1;
	mmc_card->data.nob = arg->nob;
	mmc_card->data.block_len = arg->block_len;
	mmc_card->data.sg = arg->sg;
	mmc_card->data.sg_len = arg->sg_len;

	mmc_card->llreq.cmd = &mmc_card->cmd;
	mmc_card->llreq.data = &mmc_card->data;
	
	if (arg->nob > 1) {
		if (action == MSS_READ_MEM) {
			mmc_card->cmd.opcode = MMC_READ_MULTIPLE_BLOCK;
			mmc_card->data.flags = MSS_DATA_READ | MSS_DATA_MULTI;
		}
		else {
			mmc_card->cmd.opcode = MMC_WRITE_MULTIPLE_BLOCK;
			mmc_card->data.flags = MSS_DATA_WRITE | MSS_DATA_MULTI;
		}
		
		ret = mss_send_ll_req(host, &mmc_card->llreq);
		if (!ret)
			ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
		mmc_card->state = (action == MSS_WRITE_MEM) ? CARD_STATE_RCV : CARD_STATE_DATA;
		if (ret) {
			mmc_simple_ll_req(host, mmc_card, MMC_STOP_TRANSMISSION,
				       	0, (action == MSS_WRITE_MEM) ? MSS_RESPONSE_R1B : MSS_RESPONSE_R1, 0);
			mmc_card->state = CARD_STATE_TRAN;
			if (--retries) {
				clock = host->ios.clock;
				clock = clock >> 1;
				if (clock < MMC_CARD_CLOCK_SLOW && retries == 1)
					clock = MMC_CARD_CLOCK_SLOW;
				mss_set_clock(host, clock);
				goto read_write_entry;
			}
			return ret;
		}
		ret = mmc_simple_ll_req(host, mmc_card, MMC_STOP_TRANSMISSION, 
				0, (action == MSS_WRITE_MEM) ? MSS_RESPONSE_R1B : MSS_RESPONSE_R1, 0);
		if (ret)
			return ret;
		ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
		mmc_card->state = CARD_STATE_TRAN;
		if (ret	&& (mmc_card->errno != MMC_ERROR_OUT_OF_RANGE)) 
			return ret;
	} else {
		if (action == MSS_READ_MEM) {
			mmc_card->cmd.opcode = MMC_READ_SINGLE_BLOCK;
			mmc_card->data.flags = MSS_DATA_READ;
		}
		else {
			mmc_card->cmd.opcode = MMC_WRITE_BLOCK;
			mmc_card->data.flags = MSS_DATA_WRITE;
		}
		
		ret = mss_send_ll_req(host, &mmc_card->llreq);
		if (!ret)
			ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
		if (ret) {
			if (--retries) {
				clock = host->ios.clock;
				clock = clock >> 1;
				if (clock < MMC_CARD_CLOCK_SLOW && retries == 1)
					clock = MMC_CARD_CLOCK_SLOW;
				mss_set_clock(host, clock);
				goto read_write_entry;
			}
			return ret;
		}
	}
	/* Deselect the card */
	/*mmc_simple_ll_req(host, mmc_card, MMC_SELECT_CARD, 
		0, MSS_RESPONSE_NONE, 0);
	if (ret)
		return ret;
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card);
	mmc_card->state = CARD_STATE_STBY;*/
	if (ret)
		return ret;
	result->bytes_xfered = mmc_card->data.bytes_xfered;

	return MSS_ERROR_NONE;
}

#if 0
static int mmc_lock_unlock_entry(struct mss_card *card, int action, struct mss_lock_arg *arg)
{
	struct mmc_response_r1 r1;
	struct mmc_card *mmc_card = card->prot_card;
	int ret;
	u32 status = 0;

	ret = mmc_get_status(card, &status)
	if (ret) {
		/*printk(KERN_INFO "can not get card status");
		retval = MSS_LOCK_FAILED;
		goto lock_error;*/
		return ret;
	}

	if (mmc_card->state == CARD_STATE_STBY) {
		ret = mmc_simple_ll_req(host, mmc_card, MMC_SELECT_CARD, (host->slot->rca) << 16, RESPONSE_R1B, 0);
		if (ret)
			return ret;
		ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card->state);
		if (ret)
			return ret;
	}
	mmc_card->state = CARD_STATE_TRAN;

	mmc_fix_request_block_len(arg);

	ret = mmc_simple_ll_req(host, mmc_card, MMC_SET_BLOCKLEN, arg->block_len, RESPONSE_R1, 0);
	ret = mmc_unpack_r1(&mmc_card->cmd, &r1, mmc_card->state); 
	if (ret)
		return ret;
	
/*
	mss_simple_cmd(dev, MMC_SET_BLOCKLEN, t->block_len, RESPONSE_R1, buffer);
	if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) 
		goto lock_error;
	
	cmd_backup = dev->cmd_flag;
	dev->cmd_flag = MSS_WRITE;*/
	mss_send_cmd(dev, MMC_LOCK_UNLOCK, 0, 1, t->block_len, RESPONSE_R1B, dev->io_request->buffer);
	if ((retval = mmc_unpack_r1(dev->io_request, &r1, dev->state))) { 
		dev->flags |= MSS_SLOT_FLAG_LOCK_FAILED;
		goto lock_error;
	}
	dev->cmd_flag = cmd_backup;
	
	mss_simple_cmd(dev, MMC_SEND_STATUS, (dev->slot->rca) << 16, RESPONSE_R1, buffer);
	if ((retval = mmc_unpack_r1( dev->io_request, &r1, dev->state))) 
		goto lock_error;
	
	dev->flags &= ~MSS_SLOT_FLAG_LOCK_FAILED;	

	if ( r1.status & R1_CARD_IS_LOCKED ) 
		dev->flags |= MSS_SLOT_FLAG_LOCKED;
	else 
		dev->flags &= ~MSS_SLOT_FLAG_LOCKED;

	mss_finish_io_request(dev, 1);	
	return MSS_SUCCESS;

lock_error:
	dev->flags |= MSS_SLOT_FLAG_LOCK_FAILED;
	mss_finish_io_request(dev, 0);	
	return retval; 
	return 0;
	
}
#endif

static int mmc_get_capacity(struct mss_card *card, u32 *size)
{
	int c_size = 0;
	int c_size_mult = 0;
	struct mmc_card *mmc_card = card->prot_card;
	struct mss_host *host = card->slot->host;

	if (mmc_card->access_mode == MMC_ACCESS_MODE_SECTOR && host->high_capacity)
		*size = mmc_card->ext_csd.sec_count * 512;
	else {
		c_size = mmc_card->csd.c_size;
		c_size_mult = mmc_card->csd.c_size_mult;
		*size = (c_size + 1) << (2 + c_size_mult);
	}
	dbg5("the capacity :0x%x", *size);
	return MSS_ERROR_NONE;
}


/*****************************************************************************
 *
 *   protocol driver interface functions
 *
 ****************************************************************************/

static int mmc_prot_entry(struct mss_card *card, unsigned int action, void *arg, void *result)
{
	u32 status;
	int ret;
	
	if (action != MSS_RECOGNIZE_CARD && card->card_type != MSS_MMC_CARD)
		return MSS_ERROR_WRONG_CARD_TYPE;
	switch (action) {
		case MSS_RECOGNIZE_CARD:
			ret = mmc_recognize_card(card);
			break;
		case MSS_INIT_CARD:
			ret = mmc_card_init(card);
			break;
		case MSS_READ_MEM:
		case MSS_WRITE_MEM:
			if (!arg)
				return -EINVAL;
			ret = mmc_read_write_entry(card, action, arg, result);
			break;
#if 0
		case MSS_SELECT_CARD:
		case MSS_DESELECT_CARD:
			ret = mmc_select_deselect_entry(dev);
			break;
		case MSS_SUSPEND_CARD:
			ret = mmc_suspend_entry(dev);
			break;
#endif
/*		case MSS_LOCK_UNLOCK:
			ret = mmc_lock_unlock_entry(dev);
			break;*/
		case MSS_QUERY_CARD:
			ret = mmc_get_status(card, &status);
			break;
		case MSS_GET_CAPACITY:
			ret = mmc_get_capacity(card, result);
			break;
		default:
			ret = MSS_ERROR_ACTION_UNSUPPORTED;
			//debug("Unknown protocol action!\n");
			break;
	}
	//dbg("mmc protocol entry exit, ret: %d, action: %d\n", ret, action);	
	return ret;
}

static int mmc_prot_attach_card(struct mss_card *card)
{
	struct mmc_card *mmc_card;

#define ALIGN32(x)	(((x) + 31) & (~31))
	mmc_card = kzalloc(ALIGN32(ALIGN32(sizeof(struct mmc_card))) + 512, 
			GFP_KERNEL);
	card->prot_card = mmc_card;
	if (mmc_card) {
		mmc_card->buf = (char *)ALIGN32((unsigned int)&mmc_card[1]);
		return 0;
	}
	return -ENOMEM;
}

static void mmc_prot_detach_card(struct mss_card *card)
{
	kfree(card->prot_card);
}

static int mmc_prot_get_errno(struct mss_card *card)
{
	struct mmc_card *mmc_card = card->prot_card;
	
	return mmc_card->errno;
}

#if 0
static int mmc_protocol_proc_read_device(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	struct mss_card_device *dev = (struct mss_card_device *)data;
	struct mmc_card_info *card = (struct mmc_card_info *)dev->card;
	char *p = page;
	int len = 0;

	if (!dev)
		return 0;

	p += sprintf(p, "Slot #%d\n", dev->slotid);
	p += sprintf(p, "  State %s (%d)\n", mmc_state_to_string(dev->state), dev->state);

	if ( dev->state != CARD_STATE_EMPTY ) {
		p += sprintf(p, "  Media %s\n", (dev->dev.driver ? dev->dev.driver->name : "unknown"));
		if(dev->card_type==MMC_CARD){
			p += sprintf(p, "  MMC card\n");
			p += sprintf(p, "  CID mid=%d\n", card->card_cid.mid);
			p += sprintf(p, "      oid=%d\n", card->card_cid.oid);
			p += sprintf(p, "      pnm=%s\n", card->card_cid.pnm);
			p += sprintf(p, "      prv=%d.%d\n", card->card_cid.prv>>4, card->card_cid.prv&0xf);
			p += sprintf(p, "      psn=0x%08x\n", card->card_cid.psn);
			p += sprintf(p, "      mdt=%d/%d\n", card->card_cid.mdt>>4, (card->card_cid.mdt&0xf)+1997);
    			p += sprintf(p, "  CSD csd_structure=%d\n", card->card_csd.csd_structure);
			p += sprintf(p, "      spec_vers=%d\n", card->card_csd.spec_vers);
			p += sprintf(p, "      taac=0x%02x\n", card->card_csd.taac);
			p += sprintf(p, "      nsac=0x%02x\n", card->card_csd.nsac);
			p += sprintf(p, "      tran_speed=0x%02x\n", card->card_csd.tran_speed);
			p += sprintf(p, "      ccc=0x%04x\n", card->card_csd.ccc);
			p += sprintf(p, "      read_bl_len=%d\n", card->card_csd.read_bl_len);
			p += sprintf(p, "      read_bl_partial=%d\n", card->card_csd.read_bl_partial);
			p += sprintf(p, "      write_blk_misalign=%d\n", card->card_csd.write_blk_misalign);
			p += sprintf(p, "      read_blk_misalign=%d\n", card->card_csd.read_blk_misalign);
			p += sprintf(p, "      dsr_imp=%d\n", card->card_csd.dsr_imp);
			p += sprintf(p, "      c_size=%d\n", card->card_csd.c_size);
			p += sprintf(p, "      vdd_r_curr_min=%d\n", card->card_csd.vdd_r_curr_min);
			p += sprintf(p, "      vdd_r_curr_max=%d\n", card->card_csd.vdd_r_curr_max);
			p += sprintf(p, "      vdd_w_curr_min=%d\n", card->card_csd.vdd_w_curr_min);
			p += sprintf(p, "      vdd_w_curr_max=%d\n", card->card_csd.vdd_w_curr_max);
			p += sprintf(p, "      c_size_mult=%d\n", card->card_csd.c_size_mult);
			p += sprintf(p, "      wp_grp_size=%d\n", card->card_csd.wp_grp_size);
			p += sprintf(p, "      wp_grp_enable=%d\n", card->card_csd.wp_grp_enable);
			p += sprintf(p, "      default_ecc=%d\n", card->card_csd.default_ecc);
			p += sprintf(p, "      r2w_factor=%d\n", card->card_csd.r2w_factor);
			p += sprintf(p, "      write_bl_len=%d\n", card->card_csd.write_bl_len);
			p += sprintf(p, "      write_bl_partial=%d\n", card->card_csd.write_bl_partial);
			p += sprintf(p, "      file_format_grp=%d\n", card->card_csd.file_format_grp);
			p += sprintf(p, "      copy=%d\n", card->card_csd.copy);
			p += sprintf(p, "      perm_write_protect=%d\n", card->card_csd.perm_write_protect);
			p += sprintf(p, "      tmp_write_protect=%d\n", card->card_csd.tmp_write_protect);
			p += sprintf(p, "      file_format=%d\n", card->card_csd.file_format);
			p += sprintf(p, "      ecc=%d\n", card->card_csd.ecc);	
			switch (card->card_csd.csd_structure) {
			case CSD_STRUCT_VER_1_0:
			case CSD_STRUCT_VER_1_1:
				p += sprintf(p, "      sector_size=%d\n", card->card_csd.erase.v22.sector_size);
				p += sprintf(p, "      erase_grp_size=%d\n", card->card_csd.erase.v22.erase_grp_size);
				break;
			case CSD_STRUCT_VER_1_2:
			default:
				p += sprintf(p, "      erase_grp_size=%d\n", card->card_csd.erase.v31.erase_grp_size);
				p += sprintf(p, "      erase_grp_mult=%d\n", card->card_csd.erase.v31.erase_grp_mult);
				break;
			}	  
	  	}
	}
	
  	len = (p - page) - off;
	*start = page + off;
  	return len;
}
#endif


static struct mss_prot_driver mmc_protocol = {
	.name			=	MMC_PROTOCOL,
	.prot_entry		=	mmc_prot_entry,
	.attach_card		=	mmc_prot_attach_card,
	.detach_card		=	mmc_prot_detach_card,
	.get_errno		=	mmc_prot_get_errno,
};

static int mmc_protocol_init(void)
{
	register_mss_prot_driver(&mmc_protocol);
	return 0;
}

static void mmc_protocol_exit(void)
{
	unregister_mss_prot_driver(&mmc_protocol);
}


module_init(mmc_protocol_init);
module_exit(mmc_protocol_exit);

MODULE_AUTHOR("Bridge Wu");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("MMC protocol driver");
