/* Core MMC driver functions
 *
 * Copyright (c) 2002 Hewlett-Packard Company
 *   
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *  
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *  
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Many thanks to Alessandro Rubini and Jonathan Corbet!
 *
 * Author:  Andrew Christian
 *          6 May 2002 
 */

/*
 * mss_core.c - MMC/SD/SDIO Core driver
 *
 * Copyright (C) 2006 Intel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <linux/sysctl.h>
#include <linux/suspend.h>

#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/sizes.h>
#include <asm/types.h>

#include <linux/mmc/mss_core.h>

static LIST_HEAD(mss_protocol_list);
static LIST_HEAD(mss_host_list);

/*
 * Only after card is initialized by protocol and be registed to mmc_bus, the
 * state is changed to MSS_CARD_REGISTERED.
 */
static int mmc_bus_match(struct device *dev, struct device_driver *drv)
{
	struct mss_card *card;

	card = container_of(dev, struct mss_card, dev);

	dbg(card->slot,	"bus match driver %s", drv->name);
	/* when card->state is MSS_CARD_REGISTERED,it is accepted by protocol */
	if (card->prot_driver && (card->state & MSS_CARD_REGISTERED))
		return 1;
	dbg(card->slot,	"bus match driver %s fail", drv->name);
	return 0;		
}

#if 0
static int mmc_bus_hotplug(struct device *dev, char **envp, int num_envp, char *buffer, int buffer_size)
{
	printk(KERN_INFO "*** HOT PLUG ***\n");
	return 0;
}
#endif

static int mmc_bus_suspend(struct device * dev, pm_message_t state)
{
	int ret = 0;
	struct mss_card *card;
	
	card = container_of(dev, struct mss_card, dev);

#if 0
	if (card->state & MSS_CARD_HANDLEIO)
		return -EAGAIN;
	if (card->state & MSS_CARD_SUSPENDED)
		return 0;
	
	if (dev->driver && dev->driver->suspend) {
		ret = dev->driver->suspend(dev, state, SUSPEND_DISABLE);
		if (ret == 0)
			ret = dev->driver->suspend(dev, state, SUSPEND_SAVE_STATE);
		if (ret == 0)
			ret = dev->driver->suspend(dev, state, SUSPEND_POWER_DOWN);
	}
	/* mark MSS_CARD_SUSPEND here */
	if (ret)
		card->state |= MSS_CARD_SUSPENDED;
#endif
	return ret;
}
/* 
 * card may be removed or replaced by another card from the mmc_bus when it is 
 * sleeping, and the slot may be inserted a card when it is sleeping.
 * The controller resume function need to take care about it.
 */
static int mmc_bus_resume(struct device * dev)
{
	int ret = 0;
	struct mss_card *card;
	
	card = container_of(dev, struct mss_card, dev);
	
#if 0
	/* it is new instered card or replaced card */
	if (!(card->state & MSS_CARD_SUSPENDED))
		return 0;
	
	if (dev->driver && dev->driver->resume) {
		ret = dev->driver->resume(dev, RESUME_POWER_ON);
		if (ret == 0)	
			ret = dev->driver->resume(dev, RESUME_RESTORE_STATE);
		if (ret == 0)
			ret = dev->driver->resume(dev, RESUME_ENABLE);
	}

	if (ret)
		card->state &= ~MSS_CARD_SUSPENDED;
#endif
	return ret;
}

static struct bus_type mmc_bus_type = {
	.name		=	"mmc_bus",
	.match		=	mmc_bus_match,
#if 0
	.hotplug	=	mmc_bus_hotplug,
#endif
	.suspend	=	mmc_bus_suspend,
	.resume		=	mmc_bus_resume,
};

static void mss_card_device_release(struct device *dev)
{
	struct mss_card *card = container_of(dev, struct mss_card, dev);

	kfree(card);
}

static void mss_claim_host(struct mss_host *host, struct mss_card *card)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long flags;

	spin_lock_irqsave(&host->lock, flags);
	while (host->active_card != NULL) {
		spin_unlock_irqrestore(&host->lock, flags);
		set_current_state(TASK_UNINTERRUPTIBLE);
		add_wait_queue(&host->wq, &wait);
		schedule();
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&host->wq, &wait);
		spin_lock_irqsave(&host->lock, flags);
	}
	host->active_card = card;
	spin_unlock_irqrestore(&host->lock, flags);
}

static void mss_release_host(struct mss_host* host)
{
	unsigned long flags;

	BUG_ON(host->active_card == NULL);

	spin_lock_irqsave(&host->lock, flags);
	host->active_card = NULL;
	spin_unlock_irqrestore(&host->lock, flags);

	wake_up(&host->wq);

}

int mss_card_get(struct mss_card *card)
{
	if ((card->state & MSS_CARD_REMOVING) 
			|| !(card->state & MSS_CARD_REGISTERED))
		return -ENXIO;
	if (!get_device(&card->dev))
		return -ENXIO;
	return 0;
}

void mss_card_put(struct mss_card *card)
{
	put_device(&card->dev);
}

/*
 *  finish handling a request.
 */
static void mss_finish_request(struct mss_request *req)
{
	struct mss_driver *drv;
	struct mss_card *card = req->card;

	drv = container_of(card->dev.driver, struct mss_driver, driver);
	if (drv && drv->request_done)
		drv->request_done(req);
}

/*
 * Loop all the protocol in the mss_protocol_list, and find one protocol that
 * can successful recognize and init the card.
 */
static int mss_attach_protocol(struct mss_card *card)
{
	struct list_head *item;
	struct mss_prot_driver *pdrv = NULL;
	struct mss_host *host = card->slot->host;
	int ret;

	/* loop all the protocol, and find one that match the card */
	list_for_each(item, &mss_protocol_list) {
		pdrv = list_entry(item, struct mss_prot_driver, node);
		dbg(card->slot, "try protocol:%s", pdrv->name);
		ret = pdrv->attach_card(card);
		if (ret)
			continue;
		dbg(card->slot, "protocol:%s allocate spec card", pdrv->name);
		mss_claim_host(host, card);
		ret = pdrv->prot_entry(card, MSS_RECOGNIZE_CARD, NULL, NULL);
		mss_release_host(host);
		dbg(card->slot, "protocol:%s identy ret:%d, card type:%d\n", pdrv->name, ret, card->card_type);
		if (ret)
			continue;
		switch (card->card_type) {
			case MSS_MMC_CARD:
			//case MSS_CE_ATA:
			case MSS_SD_CARD:
			case MSS_SDIO_CARD:
			case MSS_COMBO_CARD:
			//dbg("identified card type: %d", card_type);
				goto identified;
			/* 
			 * The card can be recognized, but it deos not fit the 
			 * controller.
			 */
			case MSS_UNCOMPATIBLE_CARD:
				pdrv->detach_card(card);
				return MSS_ERROR_NO_PROTOCOL;
			/* The card can not be recognized by the protocl */
			case MSS_UNKNOWN_CARD:
				pdrv->detach_card(card);
				break;	
			default:
				pdrv->detach_card(card);
				printk(KERN_WARNING "protocol driver :%s return unknown value when recognize the card\n", pdrv->name);
				break;
		}
	}
	
	return MSS_ERROR_NO_PROTOCOL;
identified:
	card->prot_driver = pdrv;
	return 0;
}

/* Initialize card by the protocol */
int mss_init_card(struct mss_card *card)
{
	int ret;
	struct mss_host *host = card->slot->host;
	
	if (!card || !card->prot_driver)
		return -EINVAL;
	mss_claim_host(host, card);
	ret = card->prot_driver->prot_entry(card, MSS_INIT_CARD, NULL, NULL);
	mss_release_host(host);

	return ret;
}

int mss_query_card(struct mss_card *card)
{	
	int ret;
	struct mss_host *host = card->slot->host;
	
	if (!card || !card->prot_driver)
		return -EINVAL;
	mss_claim_host(host, card);
	ret = card->prot_driver->prot_entry(card, MSS_QUERY_CARD, NULL, NULL);
	mss_release_host(host);

	return ret;
}

static int __mss_insert_card(struct mss_card *card) 
{
	int ret;

	dbg(card->slot, "attaching protocol");
	/* Step 1: Recognize the card */
	ret = mss_attach_protocol(card);
	if (ret)
		return ret;
	
	dbg(card->slot, "protocol%s attached", card->prot_driver->name);
	/* Step 2, initialize the card */
	ret = mss_init_card(card);
	if (ret) { 
		goto detach_prot;
	}

	dbg(card->slot, "protocol%s inited", card->prot_driver->name);
	/* Step 3, register the card to mmc bus */
	card->dev.release = mss_card_device_release;
	card->dev.parent = card->slot->host->dev;
	/* set bus_id and name */
	snprintf(&card->dev.bus_id[0], sizeof(card->dev.bus_id), "mmc%d%d", 
			card->slot->host->id, card->slot->id); 
	card->dev.bus = &mmc_bus_type; 
	
	card->state |= MSS_CARD_REGISTERED;
	ret = device_register(&card->dev); /* will call mss_card_probe */	
	if (ret) {
		ret = MSS_ERROR_REGISTER_CARD;
		card->state &= ~MSS_CARD_REGISTERED;
		goto detach_prot;
	}
	dbg(card->slot, "protocol%s registered\n", card->prot_driver->name);
	return MSS_ERROR_NONE;
	
detach_prot:
	card->prot_driver->detach_card(card);
	card->prot_driver = NULL;
	return ret;
}

/*
 * After knowing a card has been inserted into the slot, this function should 
 * be invoked. At last, load card driver in card (done by card_driver->probe).
 */
static int mss_insert_card(struct mss_slot *slot)
{
	struct mss_card * card;
	int ret;

	BUG_ON(slot->card);
	dbg(slot, "card is inserting");
	card = kzalloc(sizeof(struct mss_card), GFP_KERNEL);
	if (!card)
		return -ENOMEM;
	card->slot = slot;
	slot->card = card;

	dbg(slot, "allocate card :0x%p", card);	
	ret = __mss_insert_card(card);
	dbg(slot, "insert ret :%d", ret);
	if (ret) {
		dbg(slot, "free card");
		slot->card = NULL;
		kfree(card);
	}
	return ret;
}
#if 0
/* Delete the requests in host for the card */
static void mss_delete_requests(struct mss_host *host, struct mss_card *card)
{
	unsigned long flags;
	struct mss_request *req;
	struct list_head *pos, *n;
	
	host->action = HOST_REQ_IDLE;
	spin_lock_irqsave(&host->lock, flags);
	list_for_each_safe(pos, n, &host->req_list) {
		req = list_entry(pos, struct mss_request, node);
		if (req->card == card)
			list_del_init(pos);
		spin_unlock_irqrestore(&host->lock, flags);
		req->ret_code = MSS_ERROR_CARD_REMOVING;
		mss_finish_request(req);
		
		spin_lock_irqsave(&host->lock, flags);
	}
	spin_unlock_irqrestore(&host->lock, flags);
}
#endif

static int __mss_eject_card(struct mss_card *card)
{
	//struct mss_host *host = card->slot->host;
	
	card->state |= MSS_CARD_REMOVING;

	dbg(card->slot, "card state 0x%x", card->state);
#if 0
	/* wait for request done */	
	while (card->state & MSS_CARD_HANDLEIO)
	{
		schedule_timeout(20*HZ);
	}
	spin_lock_irqsave(&host->lock, flags);
	host->action = HOST_REQ_IDLE;
	list_for_each_safe(pos, n, &host->req_list) {
		req = list_entry(pos, struct mss_request, node);
		if (req->card == card) {
			dbg(card->slot, "delete req 0x%x", req);
			list_del_init(pos);
		}
		spin_unlock_irqrestore(&host->lock, flags);
		req->ret_code = MSS_ERROR_CARD_REMOVING;
		mss_finish_request(req);
		
		spin_lock_irqsave(&host->lock, flags);
	}
	host->action = HOST_REQ_RUN;
	spin_unlock_irqrestore(&host->lock, flags);
#endif
	if (card->state & MSS_CARD_REGISTERED) {
		device_unregister(&(card->dev));
		card->state &= ~MSS_CARD_REGISTERED;
	}
	if (card->prot_driver) {
		card->prot_driver->detach_card(card);
		card->prot_driver = NULL;	
	}
	return 0;
}

/*
 * After knowing a card has been ejected from the slot, this function should 
 * be invoked. At last, unload card driver in card(done by card_driver->remove).
 */
static int mss_eject_card(struct mss_card *card)
{
	BUG_ON(!card);

	dbg(card->slot, "eject card 0x%x", card);	
	__mss_eject_card(card);
	
	card->slot->card = NULL;
	card->slot = NULL;
	//kfree(card);
	
	return 0;
}

unsigned int mss_get_capacity(struct mss_card *card)
{
	int ret; 
	u32 cap;
	
	mss_claim_host(card->slot->host, card);
	ret = card->prot_driver->prot_entry(card, MSS_GET_CAPACITY, NULL, &cap);
	mss_release_host(card->slot->host);
	dbg5("get capcacity:0x%x, ret:%d",cap, ret); 
	if (ret)
		cap = 0;
	return cap;
}
#if 0
int mss_cancle_request(struct mss_request *req)
{
	unsigned long flags;
	struct mss_request *tmpreq;
	struct mss_host *host;
	struct list_head *pos, *n;

	host = req->card->slot->host;
	
	spin_lock_irqsave(&host->lock, flags);
	host->action = HOST_REQ_IDLE;
	list_for_each_safe(pos, n, &host->req_list) {
		tmpreq = list_entry(pos, struct mss_request, node);
		if (tmpreq == req) {
			list_del_init(pos);
			spin_unlock_irqrestore(&host->lock, flags);
			return 0;
		}
	}
	spin_unlock_irqrestore(&host->lock, flags);
	
	return -ENOENT;
}
#endif
void mss_scan_slot(struct mss_slot *slot)
{
	struct mss_card *card;
	struct mss_host *host;

	dbg(slot, "begin scan slot");	
	card = slot->card;
	host = slot->host;
	/* slot has card in it before, and the card is resuming back */
	if (card && (card->state & MSS_CARD_SUSPENDED)) { 
		dbg(slot, "suspend, card state is %d", card->state);
		/* card was ejected when it is suspended */
		if (host->ops->is_slot_empty && host->ops->is_slot_empty(slot))
			mss_eject_card(card);
		else {
			/* 
			 * if host provides is_slot_empty, and it 
			 * indicates that the card is in slot, then 
			 * we try to init it.
			 * else, we try to init it directly. Obvisouly
			 * that if there is no card in the slot, the 
			 * init will fail
			 */
			dbg(slot, "no is_slot_empty");
			if (mss_init_card(card)) {
				mss_eject_card(card);
				/* It may be replaced card, re-insert it */
				mss_insert_card(slot);
			}
		}
	}
	/* slot has card in it before, and no suspend happens */
	else if (card && (card->state & MSS_CARD_REGISTERED)) {
		dbg(slot, "register, card state is %d", card->state);
		if (host->ops->is_slot_empty) {
			dbg(slot, "has is_slot_empty");
			/* card has been ejected */
			if (host->ops->is_slot_empty(slot))
				mss_eject_card(card);
		}
		else {
			/*
			 * We try to send the status query command.
			 * If card->state has set MSS_CARD_REGISRTEED,
			 * it indicates that the card has finished 
			 * identification process, and it will response
			 * the SEND_STATUS command.
			 */
			dbg(slot, "no is_slot_empty");
			if (mss_query_card(card))
				/* Card has been ejected */
				mss_eject_card(card);
		}
	}
	/* slot has card in it, but the card is not registered */
	else if (card) {
		/* This should never be happens, because when insert fail, we will delete the card */
		BUG();
#if 0
		/* the slot is empty now */
		if (host->ops->is_slot_empty && host->ops->is_slot_empty(slot))
			mss_eject_card(card);
		else {
			/* 
			 * if the function fails, we eject the card.
			 */ 
			if (__mss_insert_card(card))
				mss_eject_card(card);
		}
#endif
	}
	/* slot has no card in it before */
	else if (!card) {
		dbg(slot, "no card in it before");
		if (host->ops->is_slot_empty) {
			dbg(slot, "has is_slot_empty");
			/* slot is not empty */
			if (!host->ops->is_slot_empty(slot))
				mss_insert_card(slot);
		}
		else {
			dbg(slot, "no is_slot_empty");
			/* 
			 * try to insert a card
			 */
			mss_insert_card(slot);
		}				
	}
	else {
		printk(KERN_ERR "Unexpected situation when scan host:%d"
				", slot:%d, card state:0x%x\n", host->id, 
				slot->id, card ? card->state : 0x0);
		BUG();
	}
}

void mss_scan_host(struct mss_host *host)
{
	struct mss_slot *slot;
	int i;
	
	for (i = 0; i < host->slot_num; i++) { 
		slot = &host->slots[i];
		mss_scan_slot(slot);
	}
}

void mss_force_card_remove(struct mss_card *card)
{
	mss_eject_card(card);
}

static void mss_wait_done(struct mss_ll_request *llreq)
{
	complete(llreq->done_data);
}

int mss_send_ll_req(struct mss_host *host, struct mss_ll_request *llreq)
{
	DECLARE_COMPLETION(complete);

	llreq->done = mss_wait_done;
	llreq->done_data = &complete;

	llreq->cmd->llreq = llreq;
	llreq->cmd->error = MSS_ERROR_NONE;
	if (llreq->data) 
		llreq->cmd->data = llreq->data;
/*	if (llreq->data && llreq->stop) {
		llreq->stop->llreq = llreq;
		llreq->stop->error = 0;
	}*/
	host->ops->request(host, llreq);
	wait_for_completion(&complete);
/*	if (llreq->cmd->error || (llreq->stop && llreq->stop-error))*/
	
	return llreq->cmd->error;
}

int mss_send_simple_ll_req(struct mss_host *host, struct mss_ll_request *llreq, struct mss_cmd *cmd, u32 opcode, u32 arg, u32 rtype, u32 flags)
{
	memset(llreq, 0x0, sizeof(struct mss_ll_request));
	memset(cmd, 0x0, sizeof(struct mss_cmd));

	cmd->opcode = opcode;
	cmd->arg = arg;
	cmd->rtype = rtype;
	cmd->flags = flags;

	llreq->cmd = cmd;

	return mss_send_ll_req(host, llreq);
}	


#if 0
/**
 *  request_handler_thread
 *  @t: no use
 *
 *  kernel thread to handle request from upper layer.
 *  thread name: mmcblockd
 */
static int request_handler_thread(void *arg)
{
	struct mss_request *req;
	struct mss_card *card;
	struct mss_host *host = (struct mss_host *)arg;
	unsigned long flags;
	int ret;
	DECLARE_WAITQUEUE(wait, current);

	/* disable all signal */
	daemonize("msshandler:%d", host->id); 
	allow_signal(SIGKILL);
	
	while (1) {
		dbg5("thread begin first");	
		add_wait_queue(&host->req_wait, &wait);
		set_current_state(TASK_INTERRUPTIBLE);
		try_to_freeze();
		dbg5("thread begin");	
		spin_lock_irqsave(&host->lock, flags);	
		if (host->action == HOST_REQ_EXIT) {
			BUG_ON(!list_empty(&host->req_list));
			spin_unlock_irqrestore(&host->lock, flags);
			break;
		}
		while (host->action == HOST_REQ_RUN && 
				!list_empty(&host->req_list)) {
			req = list_entry(host->req_list.next, 
					struct mss_request, node);
			list_del_init(host->req_list.next);
			spin_unlock_irqrestore(&host->lock, flags);

			dbg5("ready for handler");
			card = req->card;
			if (card->state & MSS_CARD_REMOVING)
				continue;
			card->state |= MSS_CARD_HANDLEIO;
			dbg5("claim host");
			mss_claim_host(host, card);
			ret = card->prot_driver->prot_entry(card, req->action,
					req->arg, req->result);
			mss_release_host(host);
			dbg5("release host");
			card->state &= ~MSS_CARD_HANDLEIO;
			req->ret_code = ret;
			if (ret)
				req->errno = card->prot_driver->get_errno(card);
			if (ret && req->retries--) {
				spin_lock_irqsave(&host->lock, flags);
				list_add_tail(&host->req_list, &req->node);
				continue;
			}
				
			mss_finish_request(req);
			spin_lock_irqsave(&host->lock, flags);	
		}
		spin_unlock_irqrestore(&host->lock, flags);
		dbg5("before schedule");
		schedule();
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&host->req_wait, &wait);
	}
		
	return 0;
}
#endif

/*
 * add controller into mss_host_list
 */
int register_mss_host(struct mss_host *host)
{
	
	list_add_tail(&host->node, &mss_host_list);
#if 0
	ret = kernel_thread(request_handler_thread, host, CLONE_KERNEL);
	if (ret < 0)
		return ret;
	host->action = HOST_REQ_RUN;
#endif
	return 0;
}

/*
 *  delete controller from mss_controller_list
 */
void unregister_mss_host(struct mss_host *host)
{
	//unsigned long flags;
	
	list_del(&host->node);
#if 0
	spin_lock_irqsave(&host->lock, flags);
	host->action = HOST_REQ_EXIT;
	spin_unlock_irqrestore(&host->lock, flags);
	wake_up_interruptible(&host->req_wait);
#endif
}

/*****************************************************************************
 *
 *   functions for protocol driver
 *
 ****************************************************************************/

/*
 * add protocol driver into mss_protocol_list
 */
int register_mss_prot_driver(struct mss_prot_driver *drv)
{
	struct list_head *item;
	struct mss_host *host;
	struct mss_slot *slot;
	int i;

	list_add(&drv->node, &mss_protocol_list); 

	list_for_each(item, &mss_host_list) {
		host = list_entry(item, struct mss_host, node);
		for (i = 0; i < host->slot_num; i++) {
			slot = &host->slots[i];
			if (!slot->card) 
				mss_scan_slot(slot);
		}
	}
	return 0;
}

/*
 * delete protocol driver from mss_protocol_list
 */
void unregister_mss_prot_driver(struct mss_prot_driver *drv)
{
	struct mss_slot *slot;
	struct list_head *item;
	struct mss_host *host;
	int i;

	list_del(&drv->node);

	list_for_each(item, &mss_host_list) {
		host = list_entry(item, struct mss_host, node);
		for (i = 0; i < host->slot_num; i++) {
			slot = &host->slots[i];
			if (slot->card && slot->card->prot_driver == drv)
				mss_eject_card(slot->card);
		}
	}

}

/*****************************************************************************
 *
 *   interfaces for card driver
 *
 ****************************************************************************/

/*
 * register card driver onto MMC bus
 */
int register_mss_driver (struct mss_driver *drv)
{
	drv->driver.bus = &mmc_bus_type;
	return driver_register(&drv->driver); /* will call card_driver->probe */
}

/*
 * unregister card driver from MMC bus
 */
void unregister_mss_driver (struct mss_driver *drv)
{
	driver_unregister(&drv->driver);
}

/*
 * enable SDIO interrupt, used by SDIO application driver
 */
void mss_set_sdio_int(struct mss_host *host, int sdio_en)
{
	struct mss_ios ios;

	memcpy(&ios, &host->ios, sizeof(ios));
	ios.sdio_int = sdio_en;
	host->ops->set_ios(host, &ios);
}

void mss_set_clock(struct mss_host *host, int clock)
{
	struct mss_ios ios;

	memcpy(&ios, &host->ios, sizeof(ios));
	if (clock > host->clk_max)
		clock = host->clk_max;
	else if (clock < host->clk_min)
		clock = host->clk_min;
	ios.clock = clock;
	host->ops->set_ios(host, &ios);
}

void mss_set_buswidth(struct mss_host *host, int buswidth)
{
	struct mss_ios ios;

	memcpy(&ios, &host->ios, sizeof(ios));
	ios.sdio_int = buswidth;
	host->ops->set_ios(host, &ios);
}

void mss_set_busmode(struct mss_host *host, int busmode)
{
	struct mss_ios ios;

	memcpy(&ios, &host->ios, sizeof(ios));
	ios.sdio_int = busmode;
	host->ops->set_ios(host, &ios);
}

int mss_send_request(struct mss_request *req)
{
	struct mss_host *host; 
	struct mss_card *card;
	unsigned long flags;
	int ret;

	if (!req->card || !req->card->slot)
		return -ENODEV;
	card = req->card;
	host = card->slot->host;
	if (req->card->state & MSS_CARD_REMOVING)
		return MSS_ERROR_CARD_REMOVING;
	card->state |= MSS_CARD_HANDLEIO;
	dbg5("claim host");
	mss_claim_host(host, card);
	ret = card->prot_driver->prot_entry(card, req->action, req->arg, 
			req->result);
	mss_release_host(host);
	dbg5("release host");
	card->state &= ~MSS_CARD_HANDLEIO;

	if (ret)
		req->errno = card->prot_driver->get_errno(card);
	
	return ret;
#if 0		
	req->retries = (req->retries > RETRIES_MAX)? RETRIES_MAX : req->retries;
	dbg5("add to list");
	spin_lock_irqsave(&host->lock, flags);
	list_add_tail(&req->node, &host->req_list);
	spin_unlock_irqrestore(&host->lock, flags);
	dbg5("wake up thread");
	wake_up_interruptible(&host->req_wait);
#endif

/*	INIT_WORK(&req->work, (void (*) (void *))mss_handle_request, req);
	queue_work(host->work_queue, &req->work);*/
}

#if 0
static void mss_handle_request(struct mss_request *req)
{
	int ret;
	struct mss_card *card = req->card;
	
	if (card->state & MSS_CARD_SUSPENDED)
		schedule_timeout(20*HZ);
	do {
		mss_claim_host(card->slot->host, card);
	
		card->state |= MSS_CARD_HANDLEIO;
		ret = card->prot_driver->prot_entry(card, req->action, req->arg, req->result);	
		card->state &= ~MSS_CARD_HANDLEIO;
		req->ret_code = ret;
		if (ret)
			req->error = card->prot_driver->get_errno();
		else {
			req->error = 0;
		}
		mss_release_host(card->slot->host, card);
	}while(req->retries-- && ret);	
	mss_finish_request(req);	
}
#endif
#if 0
/**
 *  mss_handle_io_request
 *  @io_request: mss_io_request from application driver
 *
 *  asynchronous READ/WRITE interface to application driver
 *  return value: 0: error, 1: success. This is consistence with end_request 
 *  interface for MMC/SD block device driver.
 */
int mss_send_request (struct mss_request *req)
{
	struct mss_card_device *card;
	struct mss_controller *controller;
       
	card = req->card;
	if (!card)
		return -ENODEV;
	
	controller = card->slot->controller;
	req->retries = (req->retries > RETRIES_MAX)? RETRIES_MAX : req->retries;
	list_add_tail(&req->node, &controller->request_list);
	
	/* controller will complete a request without interrupt by other request */
	/* TODO add the request to a queue of the controller */
	wake_up_interruptible(&controller->req_wait);
	return 0;
	/* 
	dbg("io_req to be handled: 0x%p, card->io_req: 0x%p", io_request, card->io_request);
	if (card->io_request) {
		printk(KERN_ERR "error! io_request in progress\n");
		return 0;
	}

	if (card->cmd_state == CARD_CMD_ACTIVE) {
		printk(KERN_ERR "error! cmd in progress\n");
		return 0;
	} 

	card->io_request = io_request;
	card->slotid = io_request->id;
	current_card = card;

	spin_lock(&req_lock);
	req_interrupt = 1;
	req_thread_action = REQ_HANDLE_REQ;
	spin_unlock(&req_lock);
	wake_up_interruptible(&req_wait);
	
	return 1;	*/
}
EXPORT_SYMBOL(mss_handle_io_request);
#endif

struct mss_host * mss_alloc_host(unsigned int slot_num, unsigned int id, unsigned int private_size)
{
	struct mss_host *host;
	int i = 0, size;

	size = sizeof(struct mss_host) + sizeof(struct mss_slot) * slot_num + private_size;
	host = (struct mss_host *)kmalloc(size, GFP_KERNEL);
	if (!host)
		return NULL;
	memset(host, 0x0, size);
	host->id = id;
	host->slot_num = slot_num;
	while(i < slot_num) {
		host->slots[i].id = i;
		host->slots[i].host = host;
		i++;
	}
	host->private = (void *)&host->slots[slot_num];
	host->active_card = NULL;
	init_waitqueue_head(&host->wq);
	/*init_waitqueue_head(&host->req_wait);
	INIT_LIST_HEAD(&host->req_list);
	*/
	return host;
}

void mss_free_host(struct mss_host *host)
{
	kfree(host);	
}

struct mss_host *mss_find_host(int id)
{
	struct list_head *pos;
	struct mss_host *host;

	list_for_each(pos, &mss_host_list) {
		host = list_entry(pos, struct mss_host, node);
		if (host->id == id)
			return host;
	}
	return NULL;
}

/*****************************************************************************
 *
 *   module init and exit functions
 *
 ****************************************************************************/

static int mss_core_driver_init(void)
{
	return bus_register(&mmc_bus_type);
}

static void mss_core_driver_exit(void)
{
	bus_unregister(&mmc_bus_type);
}

EXPORT_SYMBOL(mss_scan_slot);
EXPORT_SYMBOL(mss_scan_host);
EXPORT_SYMBOL(mss_alloc_host);
EXPORT_SYMBOL(mss_free_host);
EXPORT_SYMBOL(mss_find_host);
EXPORT_SYMBOL(mss_force_card_remove);
EXPORT_SYMBOL(register_mss_host);
EXPORT_SYMBOL(unregister_mss_host);

EXPORT_SYMBOL(register_mss_driver);
EXPORT_SYMBOL(unregister_mss_driver);
EXPORT_SYMBOL(mss_send_request);
EXPORT_SYMBOL(mss_get_capacity);

EXPORT_SYMBOL(register_mss_prot_driver);
EXPORT_SYMBOL(unregister_mss_prot_driver);
EXPORT_SYMBOL(mss_send_ll_req);
EXPORT_SYMBOL(mss_send_simple_ll_req);
EXPORT_SYMBOL(mss_set_sdio_int);
EXPORT_SYMBOL(mss_set_buswidth);
EXPORT_SYMBOL(mss_set_clock);
EXPORT_SYMBOL(mss_set_busmode);


module_init(mss_core_driver_init);
module_exit(mss_core_driver_exit);

MODULE_AUTHOR("Bridge Wu");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Core driver for MMC/SD/SDIO card");
