/*
 * pxa27x_controller.c - MMC Controller driver
 *
 * Copyright (C) 2006 Intel Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/dma-mapping.h>
#include <linux/wait.h>
#include <linux/suspend.h>
#include <linux/platform_device.h>
#include <linux/mmc/mss_core.h>

#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/sizes.h>
#include <asm/types.h>
#include <asm/dma-mapping.h>

#include <asm/arch/mainstone.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/hardware.h>
#include "pxa27x_controller.h"

#define DRIVER_NAME		"pxa2xx-mci"
#define PXA27X_HOST_BASE	(u32)(&__REG(0x41100000))
#define PXA27X_HOST_PHYBASE	(0x41100000)
#define PXA27X_HOST_CKEN	CKEN12_MMC
#define PXA27X_HOST_IRQ		IRQ_MMC
#define PXA27X_HOST_SLOTS	1

/*  enable controller INT according to mask */
static void pxa_host_enable_int(struct pxa_mss_host *pxa_host, unsigned int mask)
{
	u32 i_reg = 0;
	unsigned long flags;

	local_irq_save(flags);
	i_reg = readl(pxa_host->base + MMC_I_MASK);
	i_reg &= ~mask;
	writel(i_reg, pxa_host->base + MMC_I_MASK);
	local_irq_restore(flags);
}

/*
 *  disable controller INT according to mask
 */
static void pxa_host_disable_int(struct pxa_mss_host *pxa_host, unsigned int mask)
{
	u32 i_reg;
	unsigned long flags;

	local_irq_save(flags);
	i_reg = readl(pxa_host->base + MMC_I_MASK);
	i_reg |= mask;
	writel(i_reg, pxa_host->base + MMC_I_MASK);
	local_irq_restore(flags);
}

static void pxa_host_start_busclock(struct pxa_mss_host *pxa_host)
{
	writel(MMC_STRPCL_START_CLOCK, pxa_host->base + MMC_STRPCL);
}

static void pxa_host_stop_busclock(struct pxa_mss_host *pxa_host)
{
	u32 retries = 0xff;
	
	writel(MMC_STRPCL_STOP_CLOCK, pxa_host->base + MMC_STRPCL);
	while (retries--) {
		if (readl(pxa_host->base + MMC_I_REG) & MMC_I_REG_CLK_OFF)
			break;
		udelay(1);
	}
}

/* Stanley start!!!! */
/*
 *  setup DMA controller for data transfer 
 */
static void pxa_host_setup_data(struct pxa_mss_host *pxa_host, struct mss_data *data)
{
	u32 dcmd = 0;
	int i;
	unsigned int size;
	dma_addr_t dma;

	writel(data->nob, pxa_host->base + MMC_NUMBLK);
	writel(data->block_len, pxa_host->base + MMC_BLKLEN);
	if (data->flags & MSS_DATA_READ) { /*read*/
		dcmd = DCMD_INCTRGADDR | DCMD_FLOWSRC;
		DRCMRTXMMC = 0;
		DRCMRRXMMC = pxa_host->dma | DRCMR_MAPVLD;
	} else if (data->flags & MSS_DATA_WRITE ) { /*write*/
		dcmd = DCMD_INCSRCADDR | DCMD_FLOWTRG;
		DRCMRRXMMC = 0;
		DRCMRTXMMC = pxa_host->dma | DRCMR_MAPVLD;
	} else {
		BUG();
	}

	dcmd |= DCMD_BURST32 | DCMD_WIDTH1;

       dbg5("read to set up dma., sg_len:%d", data->sg_len);
       pxa_host->sg_idx = 0;
       pxa_host->dma_len = dma_map_sg(&pxa_host->pdev->dev, data->sg, 
			data->sg_len, (data->flags & MSS_DATA_READ) ?
                       	DMA_FROM_DEVICE : DMA_TO_DEVICE);

       for (i = 0; i < pxa_host->dma_len; i++) {
               data->bytes_xfered += sg_dma_len(&data->sg[i]);
               if (data->flags & MSS_DATA_READ) {
                       pxa_host->sg_cpu[i].dsadr =
                               pxa_host->phybase + MMC_RXFIFO;
                       pxa_host->sg_cpu[i].dtadr =
                               sg_dma_address(&data->sg[i]);
               }
               else {
                       pxa_host->sg_cpu[i].dsadr =
                               sg_dma_address(&data->sg[i]);
                       pxa_host->sg_cpu[i].dtadr =
                               pxa_host->phybase + MMC_TXFIFO;
               }
               pxa_host->sg_cpu[i].dcmd = dcmd | sg_dma_len(&data->sg[i]);
               pxa_host->sg_cpu[i].ddadr = pxa_host->sg_dma + (i + 1) *
                                       sizeof(struct pxa_dma_desc);
               dbg5("sg:%d, dsadr:0x%x, dtadr:0x%x, dcmd:0x%x, ddadr:0x%x", i, pxa_host->sg_cpu[i].dsadr, pxa_host->sg_cpu[i].dtadr, pxa_host->sg_cpu[i].dcmd, pxa_host->sg_cpu[i].ddadr);
       }
       pxa_host->sg_cpu[pxa_host->dma_len - 1].ddadr = DDADR_STOP;
       pxa_host->sg_cpu[pxa_host->dma_len - 1].dcmd |= DCMD_ENDIRQEN;
       wmb();

#if 0
	pxa_host->buffer_size = data->block_len * data->nob;
	dma = pxa_host->dma_buf_addr;
	for (i = 0, size = pxa_host->buffer_size; size; i++) {
		u32 len = 32;
		if ( size < 32 ) {
			len = size;
		}
		if (data->flags & MSS_DATA_READ) {
			pxa_host->dma_desc[i].dsadr = pxa_host->phybase + MMC_RXFIFO;
			pxa_host->dma_desc[i].dtadr = dma;
		} else {
			pxa_host->dma_desc[i].dsadr = dma;
			pxa_host->dma_desc[i].dtadr = pxa_host->phybase + MMC_TXFIFO;
		}
		pxa_host->dma_desc[i].dcmd = dcmd | len;
		dma += len;
		size -= len;

		if (size) {
			pxa_host->dma_desc[i].ddadr = pxa_host->dma_desc_addr + (i + 1) * sizeof(pxa_dma_desc);
		} else {
			pxa_host->dma_desc[i].dcmd  = (pxa_host->dma_desc[i].dcmd) | DCMD_ENDIRQEN ;
			pxa_host->dma_desc[i].ddadr = DDADR_STOP;
		}
	}
#endif
}

/*
 *  read response of MMC/SD/SDIO controller.
 */
static void pxa_host_get_response(struct pxa_mss_host *pxa_host, struct mss_cmd *cmd)
{
	int i;
	u32 mmc_res;

	for (i = 0; i < PXA_MSS_MAX_RESPONSE_SIZE; i = i + 2) {
		mmc_res = __REG(0x4110003c);
		if (i < MSS_MAX_RESPONSE_SIZE) {
			cmd->response[i] = mmc_res  >> 8;
			cmd->response[i + 1] = mmc_res & 0xff;
		}
	}

#if defined(CONFIG_MMC_DEBUG)
	printk("host%d, response:", pxa_host->host->id);
	for (i = 0; i < MSS_MAX_RESPONSE_SIZE; i++)
		printk("0x%x, ", cmd->response[i]);
	printk("\n");
#endif
}

/*
 *  set io_request result according to error status in MMC/SD/SDIO controller 
 *  status reg.
 */
static void pxa_host_set_error(struct mss_cmd *cmd, u32 stat)
{

	if (stat & MMC_STAT_RESP_TIMEOUT ) 
		cmd->error = MSS_ERROR_TIMEOUT;
	else if (stat & MMC_STAT_RESP_CRC) 
		cmd->error = MSS_ERROR_CRC;
	else if (stat & MMC_STAT_FLASH_ERR)
		cmd->error = MSS_ERROR_FLASH;
	else if (stat & MMC_STAT_READ_TIMEOUT)
		cmd->error = MSS_ERROR_TIMEOUT;
	else if (stat & (MMC_STAT_READ_CRC | MMC_STAT_WRITE_CRC))
		cmd->error = MSS_ERROR_CRC;
	dbg5("get error %d", cmd->error);
}

/*
 *  return value: 0 -- not write-protected, 1 -- write-protected
 */
static int pxa_mss_slot_is_wp(struct mss_slot *slot)
{
	if (MST_MSCRD & MST_MSCRD_nMMC_WP)
		return 0;
	else 
		return 1;
}

/*
 *  return value: 0 -- inserted, 1 -- empty.
 */
static int pxa_mss_slot_is_empty(struct mss_slot *slot)
{
	if (MST_MSCRD & MST_MSCRD_nMMC_CD)
		return 1;
	else
		return 0;
}

static void pxa_mss_set_ios(struct mss_host *host, struct mss_ios *ios)
{
	unsigned int clockrate;
	struct pxa_mss_host *pxa_host;
       
	pxa_host = host->private;
	
	dbg5("clock now %d, to set %d, CLKRT:%x", host->ios.clock, ios->clock, readl(pxa_host->base + MMC_CLKRT));
	if (ios->clock != host->ios.clock) {
#ifdef CONFIG_CPU_MONAHANS_L
		if (ios->clock >= 26000000) {
			host->ios.clock = 26000000;
			clockrate = CLKRT_26MHZ;
		}
		else if (ios->clock >= 19500000)
#else
		if (ios->clock >= 19500000)
#endif
		{
			host->ios.clock = 19500000;
			clockrate = CLKRT_19_5MHZ;
		}
		else if (ios->clock >= 9750000) {
			host->ios.clock = 9750000;
			clockrate = CLKRT_9_75MHZ;
		}
		else if (ios->clock >= 4875000) {
			host->ios.clock = 4875000;
			clockrate = CLKRT_4_88MHZ;
		}
		else if (ios->clock >= 2437500) {
			host->ios.clock = 2437500;
			clockrate = CLKRT_2_44MHZ;
		}	
		else if (ios->clock >= 1218750) {
			host->ios.clock = 1218750;
			clockrate = CLKRT_1_22MHZ;     
		}
		else if (ios->clock >= 609375){
			host->ios.clock = 609375;
			clockrate = CLKRT_0_609MHZ; 
		}	
		else {
			host->ios.clock = 304000;
			clockrate = CLKRT_0_304MHZ; 
		}
		dbg5("--------read to to set %d", clockrate);
		pxa_host_stop_busclock(pxa_host);
		writel(clockrate, pxa_host->base + MMC_CLKRT);
		pxa_host_start_busclock(pxa_host);
	}
	host->ios.sdio_int = ios->sdio_int;
	if (host->ios.sdio_int == MSS_SDIO_INT_EN)
		pxa_host_enable_int(pxa_host, MMC_I_MASK_SDIO_INT);
	else
		pxa_host_disable_int(pxa_host, MMC_I_MASK_SDIO_INT);
	host->ios.bus_width = ios->bus_width;
	if (host->ios.bus_width > host->bus_width)
		host->ios.bus_width = host->bus_width;

}

/*
 *  send command by setting CMDAT reg of MMC/SD/SDIO controller.
 *  Note: to hide info about SD/SDIO.
 */
static void pxa_mss_handle_request(struct mss_host *host, struct mss_ll_request *llreq)
{
	struct pxa_mss_host *pxa_host = host->private;
	struct mss_cmd *cmd;
	struct mss_card *card = host->active_card;
	u32 cmd_dat = 0;

	dbg4(card->slot, "pxammc acitive slot %x", (pxa_host->active_slot)?pxa_host->active_slot->id:-1);
	if (pxa_host->active_slot != card->slot) {
		dbg4(card->slot, "pxammc change acitive slot to %d", card->slot->id);
		/* pxa_mss_slot_select(card->slot); */
		pxa_host->active_slot = card->slot;
	}
	pxa_host->cmd = llreq->cmd;
	pxa_host->data = llreq->data;
	pxa_host->llreq = llreq;
	cmd = pxa_host->cmd;

	dbg4(card->slot, "cmd flag 0x%x", cmd->flags);
	if (cmd->flags & MSS_CMD_INIT)
		cmd_dat |= MMC_CMDAT_INIT;
	if (cmd->flags & MSS_CMD_STOP)
		cmd_dat |= MMC_CMDAT_STOP;
	if (cmd->flags & MSS_CMD_SDIO_EN)
		cmd_dat |= MMC_CMDAT_SDIO_INT;
	if (llreq->data) {
		dbg4(card->slot, "data flags 0x%x", llreq->data->flags);
		//BUG_ON(!llreq->data->buffer);
		if (llreq->data->flags & MSS_DATA_WRITE)
			cmd_dat |= MMC_CMDAT_WRITE;
		pxa_host_setup_data(pxa_host, llreq->data);
		cmd_dat |= (MMC_CMDAT_DATA | MMC_CMDAT_DMA);
		cmd_dat &= ~MMC_CMDAT_BUSY;
		writel(llreq->data->nob, pxa_host->base + MMC_NUMBLK);
		writel(llreq->data->block_len, pxa_host->base + MMC_BLKLEN);
		dbg4(card->slot, "data nob %d, block_len %d", llreq->data->nob, llreq->data->block_len);
		if (llreq->data->flags & MSS_DATA_STREAM)
			cmd_dat |= MMC_CMDAT_STREAM;

	}
	
	dbg4(card->slot, "cmd rtype %d", cmd->rtype);
	switch (cmd->rtype) {
		case MSS_RESPONSE_NONE:
			cmd_dat |= MMC_CMDAT_RESP_FMT_NONE;
			break;
		case MSS_RESPONSE_R1B:
			cmd_dat |= MMC_CMDAT_BUSY;
			/* Fall through */
		case MSS_RESPONSE_R1:
		case MSS_RESPONSE_R4:
		case MSS_RESPONSE_R5:
		case MSS_RESPONSE_R6:
			cmd_dat |= MMC_CMDAT_RESP_FMT_R1;
			break;
		case MSS_RESPONSE_R2_CID:
		case MSS_RESPONSE_R2_CSD:
			cmd_dat |= MMC_CMDAT_RESP_FMT_R2;
			break;
		case MSS_RESPONSE_R3:
			cmd_dat |= MMC_CMDAT_RESP_FMT_R3;
			break;
	}

	dbg4(card->slot, "card bus width %d", card->bus_width);
	if (card->bus_width == MSS_BUSWIDTH_4BIT)
		cmd_dat |= MMC_CMDAT_4DAT;
		
	dbg4(card->slot, "cmd opcode 0x%x, arg 0x%x", cmd->opcode, cmd->arg);
	dbg4(card->slot, "CMDAT 0x%x", cmd_dat);
	writel(cmd->opcode, pxa_host->base + MMC_CMD);
	writel(cmd->arg >> 16, pxa_host->base + MMC_ARGH);
	writel(cmd->arg & 0xffff, pxa_host->base + MMC_ARGL);
	writel(cmd_dat, pxa_host->base + MMC_CMDAT);
	dbg4(card->slot, "Read out MMC_CMD 0x%x",readl(pxa_host->base+MMC_CMD));
	dbg4(card->slot,"Read out MMC_ARGH 0x%x",readl(pxa_host->base+MMC_ARGH));
	dbg4(card->slot,"Read out MMC_ARGL 0x%x",readl(pxa_host->base+MMC_ARGL));
	dbg4(card->slot,"Read out MMC_CMDAT 0x%x",readl(pxa_host->base+MMC_CMDAT));
 	
	if (llreq->data && llreq->data->sg_len) {
		DDADR(pxa_host->dma) = pxa_host->sg_dma;
		/* start DMA */
		DCSR(pxa_host->dma) |= DCSR_RUN; 
		dbg5("enable dma %d", pxa_host->dma);
	}

	pxa_host_enable_int(pxa_host, MMC_I_MASK_END_CMD_RES | MMC_I_MASK_RES_ERR);	
}

/*
 * DMA irq handler. devid represent pxa_mss_controller_device for dma interrupt
 * is to pxa_controller
 */
static void pxa_host_dma_irq(int dma, void *devid, struct pt_regs *regs)
{
	unsigned int dcsr;
	struct pxa_mss_host *pxa_host;
	struct mss_card *card;
	struct mss_cmd *cmd;
	struct mss_data *data;

	pxa_host = (struct pxa_mss_host *)devid;
	card = pxa_host->host->active_card;
	data = pxa_host->data;
	if (!card) {
		printk(KERN_ERR "Can not find card\n");
		BUG();
	}
	cmd = pxa_host->cmd;

	dcsr = DCSR(dma);
	DCSR(dma) &= ~DCSR_STOPIRQEN;	
	dbg5("dma int dcsr:0x%x", dcsr);
	if (dcsr & DCSR_BUSERR) {
		dbg5("DCSR_BUSEER");
		cmd->error = MSS_ERROR_DMA;
		data->bytes_xfered = 0;
		pxa_host_disable_int(pxa_host, MMC_I_MASK_DATA_TRAN_DONE | 
				MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT);
		pxa_host->llreq->done(pxa_host->llreq);
	}
       else if (dcsr & DCSR_STOPSTATE) {
               dbg5("sg index:%d, dma_len:%d", pxa_host->sg_idx, pxa_host->dma_len);
               if (data && (data->flags & (MSS_DATA_WRITE | MSS_DATA_READ))) {
                       pxa_host_enable_int(pxa_host, MMC_I_MASK_DATA_TRAN_DONE
                                       | MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT);
               } else {
                       dbg5("ERROR in SLOT_DMA_IRQ, neither WRITE nor READ or sg_idx exceed dma_len");
                       dump_stack();
               }
	}	
	return;	
}

/*
 *  slot INT routine for commands with R1B response type
 */
static int pxa_host_r1b_irq(struct pxa_mss_host *pxa_host, u32  ireg, u32 stat, u32 mask)
{
	struct mss_host *host;
	struct mss_cmd *cmd;

	host = pxa_host->host;
	cmd = pxa_host->cmd;
	
	if ((ireg & MMC_I_REG_END_CMD_RES) && (!(mask & MMC_I_MASK_END_CMD_RES)) ) {
		pxa_host_disable_int(pxa_host, MMC_I_MASK_END_CMD_RES | MMC_I_MASK_RES_ERR);
		pxa_host_get_response(pxa_host, cmd);
		if (ireg & MMC_I_REG_RES_ERR) {
			pxa_host_set_error(cmd, stat);
			dbg5("in RES_ERROR, prepare to cmd_complete\n");
			pxa_host->llreq->done(pxa_host->llreq);
			return 1;
		}		
		pxa_host_enable_int(pxa_host, MMC_I_MASK_PRG_DONE);
		return 1;
	}
	if ((ireg & MMC_I_REG_PRG_DONE) && (!(mask & MMC_I_MASK_PRG_DONE)) ) {
#if 0
		pxa_host_get_response(pxa_host, cmd);
#endif
		pxa_host_disable_int(pxa_host, MMC_I_MASK_PRG_DONE);
		dbg5("in PRG_DONE, prepare to cmd_complete\n");
		pxa_host->llreq->done(pxa_host->llreq); 
	}
	return 1;
}

/*
 *  slot INT routine for write commands with data buffer
 */
static int pxa_host_write_irq(struct pxa_mss_host *pxa_host, u32  ireg, u32 stat, u32 mask)
{
	struct mss_host *host;
	struct mss_cmd *cmd;

	host = pxa_host->host;
	cmd = pxa_host->cmd;
	
	if ((ireg & MMC_I_REG_END_CMD_RES) && (!(mask & MMC_I_MASK_END_CMD_RES)) ) {
		pxa_host_disable_int(pxa_host, MMC_I_MASK_END_CMD_RES | MMC_I_MASK_RES_ERR);
		if (ireg & MMC_I_REG_RES_ERR) {
			pxa_host_set_error(cmd, stat);
			pxa_host->llreq->done(pxa_host->llreq);
			return 1;
		}	
		pxa_host_get_response(pxa_host, cmd);
		return 1;
	}

	if ((ireg & (MMC_I_REG_DAT_ERR | MMC_I_REG_TINT)) 
			&& (!(mask & (MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT)))) {
		stat = readl(pxa_host->base + MMC_STAT);
		pxa_host_disable_int(pxa_host, MMC_I_MASK_DATA_TRAN_DONE | MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT);
		pxa_host_set_error(cmd, stat);
		pxa_host->llreq->done(pxa_host->llreq);
		return 1;
	}	

#if 0
	DDADR(pxa_controller->dma) = pxa_controller->dma_desc_addr;
	DCSR(pxa_controller->dma) |= DCSR_RUN; /* start dma transfer */
#endif
	/* wait for DMA interrupt, DATA WRITE done once DATA_TRAN_DONE interrupt occurred */
	if ( (ireg & MMC_I_REG_DATA_TRAN_DONE) && (!(mask & MMC_I_MASK_DATA_TRAN_DONE)) ) {
		/* Note: must not disable DATA_ERROR interrupt, for it may be occured again!!! */
		pxa_host_disable_int(pxa_host, MMC_I_MASK_DATA_TRAN_DONE | MMC_I_MASK_TINT);
		pxa_host_enable_int(pxa_host, MMC_I_MASK_PRG_DONE); 
		return 1;
	}	
	
	if ( (ireg & MMC_I_REG_PRG_DONE) && (!(mask & MMC_I_MASK_PRG_DONE)) ) {
		pxa_host_get_response(pxa_host, cmd);
		pxa_host_disable_int(pxa_host, MMC_I_MASK_PRG_DONE);
		pxa_host->llreq->done(pxa_host->llreq); 
	}
	return 1;
}

/**
 *  pxa_slot_read_irq
 *  @card: mss_card_device
 *  @ireg: value of MMCx_I_REG
 *  @stat: value of MMCx_STAT
 *  @mask: value of MMCx_I_MASK
 *
 *  slot INT routine for read commands with data buffer
 */
static int pxa_host_read_irq(struct pxa_mss_host *pxa_host, u32  ireg, u32 stat, u32 mask)
{
	struct mss_host *host;
	struct mss_cmd *cmd;

	host = pxa_host->host;
	cmd = pxa_host->cmd;

	if ((ireg & MMC_I_REG_END_CMD_RES) && (!(mask & MMC_I_MASK_END_CMD_RES))) {
		pxa_host_disable_int(pxa_host, MMC_I_MASK_END_CMD_RES | MMC_I_MASK_RES_ERR);
		if (ireg & MMC_I_REG_RES_ERR) {
			pxa_host_set_error(cmd, stat);
			pxa_host->llreq->done(pxa_host->llreq);
			return 1;
	  	}
	  	pxa_host_get_response(pxa_host, cmd);
		/* DATA_TRAN_DONE is not enabled now */
		pxa_host_enable_int(pxa_host, MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT);
		return 1;
	}

#if 0
	DDADR(pxa_controller->dma) = pxa_controller->dma_desc_addr;
	DCSR(pxa_controller->dma) |= DCSR_RUN; /* start dma transfer */
#endif
	if ((ireg & (MMC_I_REG_DAT_ERR | MMC_I_REG_TINT)) && (!(mask & (MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT)))) {
		stat = readl(pxa_host->base + MMC_STAT);
		pxa_host_disable_int(pxa_host, MMC_I_MASK_DATA_TRAN_DONE | MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT);
		pxa_host_set_error(cmd, stat);
		dbg5("DATA/TIMEOUT error.  I_REG: 0x%x, STAT: 0x%x, I_MASK: 0x%x. prepare to cmd_complete", readl(pxa_host->base + MMC_I_REG),readl(pxa_host->base + MMC_STAT),readl(pxa_host->base + MMC_I_MASK)); 
		pxa_host->llreq->done(pxa_host->llreq);
	}

	/* wait for DMA interrupt, DATA READ done once DATA_TRAN_DONE interrupt occurred */
	if ((ireg & MMC_I_REG_DATA_TRAN_DONE) && (!(mask & MMC_I_MASK_DATA_TRAN_DONE))) {
#if 0
		pxa_host_get_response(pxa_host, cmd);
#endif
		pxa_host_disable_int(pxa_host, MMC_I_MASK_DATA_TRAN_DONE | MMC_I_MASK_DAT_ERR | MMC_I_MASK_TINT);
		dbg5("in DATA_TRAN_DONE, prepare to cmd_complete");
		pxa_host->llreq->done(pxa_host->llreq); 
		return 1;
	}
	
	return 1;
}

/**
 *  pxa_slot_other_irq
 *  @card: mss_card_device
 *  @ireg: value of MMCx_I_REG
 *  @stat: value of MMCx_STAT
 *  @mask: value of MMCx_I_MASK
 *
 *  slot INT routine for commands other than R1B, write and read commands
 */
static int pxa_host_other_irq(struct pxa_mss_host *pxa_host, u32  ireg, u32 stat, u32 mask)
{
	struct mss_cmd *cmd = pxa_host->cmd;

	dbg5("I_REG: 0x%x, STAT: 0x%x, MASK: 0x%x\n", (u32)ireg, (u32)stat, (u32)mask);
		
	if ((ireg & MMC_I_REG_RES_ERR) && (!(mask & MMC_I_MASK_RES_ERR))){
		pxa_host_disable_int(pxa_host, MMC_I_MASK_END_CMD_RES | MMC_I_MASK_RES_ERR);
		pxa_host_get_response(pxa_host, cmd);
		pxa_host_set_error(cmd, stat);
		dbg5("in RES_ERROR, prepare to cmd_complete\n");
		pxa_host->llreq->done(pxa_host->llreq);
		return 1;
	} else if((ireg & MMC_I_REG_END_CMD_RES) && (!(mask & MMC_I_MASK_END_CMD_RES))) {
		pxa_host_disable_int(pxa_host, MMC_I_MASK_END_CMD_RES | MMC_I_MASK_RES_ERR);
		pxa_host_get_response(pxa_host, cmd);
		pxa_host->llreq->done(pxa_host->llreq);
	}
	
	dbg5("exit from OTHER\n");	
	return 1;
}

/**
 *  pxa_host_irq
 *  @irq: IRQ number
 *  @devid: pointer to pxa_mss_controller_device
 *  @regs: interrupt context
 *
 *  response interrupt and sdio interrupt handler. 
 *  devid represent pxa_mss_controller_device because interrupt is to pxa_controller
 */
static irqreturn_t pxa_host_irq(int irq, void *devid, struct pt_regs *regs)
{
	struct pxa_mss_host *pxa_host;
	struct mss_cmd *cmd;
	struct mss_data *data;
	struct mss_driver *drv;
	struct mss_card *card;
	u32 ireg, stat, mask;

	pxa_host = (struct pxa_mss_host *)devid;
	cmd = pxa_host->cmd;
	data = pxa_host->data;
	card = pxa_host->host->active_card;

	ireg = readl(pxa_host->base + MMC_I_REG);
	stat = readl(pxa_host->base + MMC_STAT);
	mask = readl(pxa_host->base + MMC_I_MASK);

	if (data) {
		//BUG_ON(!data->buffer);
		if (data->flags & MSS_DATA_WRITE) {
			pxa_host_write_irq(pxa_host, ireg, stat, mask);
		}
		if (data->flags & MSS_DATA_READ) {
			pxa_host_read_irq(pxa_host, ireg, stat, mask);
		}
	}
	else if ((cmd->rtype == MSS_RESPONSE_R1B) && (!data || !data->sg_len))
		pxa_host_r1b_irq(pxa_host, ireg, stat, mask);
	else
		pxa_host_other_irq(pxa_host, ireg, stat, mask);
		
	/* handle sdio interrupt, wakeup sdio int IRQ handler thread */
	if ((ireg & MMC_I_REG_SDIO_INT) && (!(mask & MMC_I_MASK_SDIO_INT))) {
		//dbg2("SDIO interrupt ocurred");	

		pxa_host_disable_int(pxa_host, MMC_I_MASK_SDIO_INT);
		drv = container_of(card->dev.driver, struct mss_driver, driver);
		INIT_WORK(&card->slot->sdio_int,
				(void (*)(void *))drv->sdio_int_handler, card);
		queue_work(pxa_host->work_queue, &card->slot->sdio_int);
	}

	return IRQ_HANDLED;
}

/**
 *  pxa_slot_cd_irq
 *  @irq: IRQ number
 *  @devid: pointer to mss_slot
 *  @regs: interrupt context
 *
 *  hot-plug interrupt handler. devid represent slot because hot-plug is slot-specific. 
 */
static irqreturn_t pxa_slot_cd_irq(int irq, void *devid, struct pt_regs *regs)
{
	struct mss_slot *slot = (struct mss_slot *)devid;
	struct pxa_mss_host *pxa_host = slot->host->private;

	dbg4(slot, "card detect irq");
	/* delay 200ms for card stability */
	queue_delayed_work(pxa_host->work_queue, &slot->card_detect, HZ/5);
	
	return IRQ_HANDLED;
}

/* init slot, request IRQ for slot, register mss_slot to core, initialize card 
 * if inserted 
 */
static int pxa_slot_init(struct pxa_mss_host *pxa_host, int slotid)
{
	struct mss_slot *slot;
	struct mss_host *host = pxa_host->host;
	int ret = 0;
	
	slot = &host->slots[slotid];

	/* stanley temporary solution */
	MST_INTMSKENA |= 1;
	ret = request_irq(MAINSTONE_MMC_IRQ, pxa_slot_cd_irq, 0, "MMC card detection", (void *)slot);
	if (ret) {
		printk(KERN_ERR "MMC:slot%d, request irq %d fails, ret: %d\n", slotid, MAINSTONE_MMC_IRQ, ret);
		return ret;
	}

	INIT_WORK(&slot->card_detect, (void (*)(void *))mss_scan_slot, slot);
	mss_scan_slot(slot);
	
	return ret;
}

/**
 *  cleanup a slot, unregister mss_slot form core, free IRQ of slot, 
 *  free mss_slot 
 */
static void pxa_slot_exit(struct mss_slot *slot)
{
	if (slot->card)
		mss_force_card_remove(slot->card);
	free_irq(MAINSTONE_MMC_IRQ, slot);
}
static void pxa_platform_init( void )
{
	MST_MSCWR1 |= MST_MSCWR1_MMC_ON;
	MST_MSCWR1 &= ~MST_MSCWR1_MS_SEL;
	mdelay(10);
}

static void pxa_platform_shutdown( void )
{
	MST_MSCWR1 &= ~MST_MSCWR1_MMC_ON;
}

/* Only support 1 MMC/SD/SDIO card -stanley */
static void pxa_mss_host_init(struct pxa_mss_host *pxa_host)
{
	pxa_host->base = PXA27X_HOST_BASE;
	pxa_host->phybase = PXA27X_HOST_PHYBASE;
	pxa_host->cken = PXA27X_HOST_CKEN;
	pxa_set_cken(pxa_host->cken, 1);

        pxa_gpio_mode(GPIO32_MMCCLK_MD);
	pxa_gpio_mode(GPIO112_MMCCMD_MD);
	pxa_gpio_mode(GPIO92_MMCDAT0_MD);
	pxa_gpio_mode(GPIO109_MMCDAT1_MD);
	pxa_gpio_mode(GPIO110_MMCDAT2_MD);
	pxa_gpio_mode(GPIO111_MMCDAT3_MD);

	writel(CLKRT_0_304MHZ, pxa_host->base + MMC_CLKRT);
	writel(RES_TIMEOUT_MAX, pxa_host->base + MMC_RESTO);
	writel(RD_TIMEOUT_MAX, pxa_host->base + MMC_RDTO);
	writel(MMC_MASK_ALL, pxa_host->base + MMC_I_MASK);
	writel(0, pxa_host->base + MMC_SPI);
	writel(MMC_STRPCL_START_CLOCK, pxa_host->base + MMC_STRPCL);


	/* a nasty hack */
	pxa_platform_init();
#if 0
	dbg5("*********CLKRT:%x", readl(pxa_host->base + MMC_CLKRT));
	dbg5("*********CLKRT:%x", readl(pxa_host->base + MMC_CLKRT));
	dbg5("*********CLKRT:%x", readl(pxa_host->base + MMC_CLKRT));
#endif
}

static struct mss_host_ops pxa_mss_host_ops = {
	.request	= pxa_mss_handle_request,
	.set_ios	= pxa_mss_set_ios,
	.is_slot_empty	= pxa_mss_slot_is_empty,
	.is_slot_wp	= pxa_mss_slot_is_wp,
};

/**
  *  pxa_mss_controller_probe
  *  @dev:	device
  *  
  *  init controller, request CTROLLER IRQ, request DMA IRQ, create controller
  *  threads, invoke slot_init
  */
static int pxa_mss_host_probe(struct platform_device *pdev)
{
	int ret = 0, i = 0;
	struct pxa_mss_host *pxa_host;
	struct mss_host *host;
	struct resource *r;
	int irq;

	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	irq = platform_get_irq(pdev, 0);
	if (!r || irq < 0)
		return -ENXIO;

#if 0
	if (pdev->id > PXA_MMC_MAX)
		return -ENODEV;
#endif

	host = mss_alloc_host(PXA27X_HOST_SLOTS, pdev->id, sizeof(struct pxa_mss_host));
	pxa_host = (struct pxa_mss_host *)host->private;
	pxa_host->host = host;
	pxa_host->irq = irq;

	host->dev = &pdev->dev;
	host->ops = &pxa_mss_host_ops;
	host->vdd = host->ios.vdd = MSS_VDD_32_33;
        host->bus_width = MSS_BUSWIDTH_4BIT;
        host->ios.bus_width = MSS_BUSWIDTH_1BIT;
	host->clk_min = 304000;
	host->clk_max = 26000000;
        host->sd_spec = MSS_SD_SPEC_1X;
        host->mmc_spec = MSS_MMC_SPEC_31;
        host->sdio_spec = MSS_SDIO_SPEC_11;
        host->max_phys_segs = 32;
        host->max_hw_segs = 32;
        host->max_sectors = 256;
        host->max_seg_size = 0x1000;

	ret = register_mss_host(host);
	if (ret) {
		printk(KERN_ERR "register_mss_host error! ret: %d.\n", ret);
		goto free_host;
	}
	pxa_mss_host_init(pxa_host);

	/* request controller IRQ */
	ret = request_irq(pxa_host->irq, pxa_host_irq, 0, \
		"PXA27x MMC/SD/SDIO Controller", (void *)pxa_host);
	if (ret < 0) {
		printk(KERN_ERR "MMC%d request_irq error! ret: %d.\n", pdev->id, ret);
		goto unregister_host;
	}
	/* request DMA IRQ */
	pxa_host->dma = pxa_request_dma("PXA27x", DMA_PRIO_LOW, 
				pxa_host_dma_irq, (void *)pxa_host);
	if (pxa_host->dma < 0) {
		ret = -EBUSY;
		printk(KERN_ERR "MMC%d request_dma error! ret: %d.\n", pdev->id, ret);
		goto free_host_irq;
	}

	pxa_host->sg_cpu = dma_alloc_coherent(&pdev->dev, PAGE_SIZE,
				&pxa_host->sg_dma, GFP_KERNEL);
	if (!pxa_host->sg_cpu) {
		ret = -ENOMEM;
		goto free_dma_irq;
	}

	pxa_host->work_queue = create_workqueue("pxa_host");
	for(i = 0; i < PXA27X_HOST_SLOTS; i++) {
		pxa_slot_init(pxa_host, i); 
	}	
	printk("pxa mss slot initialized.\n");
	
	return 0;
free_dma_irq:
	/* release DMA IRQ */
	pxa_free_dma(pxa_host->dma);

free_host_irq:
	/* release controller IRQ */
	free_irq(pxa_host->irq, pxa_host);

unregister_host:
	unregister_mss_host(host);

free_host:
	mss_free_host(host);
	return ret;
}


/**
 *  pxa_mss_controller_remove
 *  @dev: device
 *  
 *  cleanup controller, free CTROLLER IRQ, free DMA IRQ, destroy controller
 *  threads, invoke slot_exit
 */
static int pxa_mss_host_remove(struct platform_device *pdev)
{
	int ret = 0, i;
	struct mss_host *host;
	struct pxa_mss_host *pxa_host;

	host = mss_find_host(pdev->id);
	if (!host)
		return -ENODEV;

	pxa_host = (struct pxa_mss_host *)host->private;

	free_irq(pxa_host->irq, pxa_host);
	pxa_free_dma(pxa_host->dma);
	dma_free_coherent(&pdev->dev, PAGE_SIZE, pxa_host->sg_cpu, pxa_host->sg_dma);
	for (i = 0; i < host->slot_num; i++) {
		pxa_slot_exit(&host->slots[i]);
	}

	unregister_mss_host(host);
	mss_free_host(host);

	return ret;
}

static void pxa_mss_host_shutdown(struct platform_device *pdev)
{
	pxa_platform_shutdown();
	return;
}

/*****************************************************************************
 *
 *   object instances
 *
 ****************************************************************************/
static struct platform_driver pxamci_driver = {
	.probe		= 	pxa_mss_host_probe,
	.remove		=	pxa_mss_host_remove,
	.shutdown	=	pxa_mss_host_shutdown,
	.suspend	=	NULL,	
	.resume		=	NULL,	
	.driver		=	{
		.name = DRIVER_NAME,
	},
};

static int pxa_mss_host_driver_init(void)
{
	printk("mss: pxamci host registered.\n");
	return platform_driver_register(&pxamci_driver);
}

static void pxa_mss_host_driver_exit(void)
{
	platform_driver_unregister(&pxamci_driver);
}

module_init(pxa_mss_host_driver_init);
module_exit(pxa_mss_host_driver_exit);

MODULE_AUTHOR("Bridge Wu");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Controller driver for MMC/SD/SDIO card");
